#ifndef _FILESYS_H_
#define _FILESYS_H_



#define DEVICE_ID_MAX 4
#define FILESYSTEM_ID_MAX 4



#define offsetof(type, member) ((unsigned long) &(((type*)0)->member))



// A Filesystem consists of an API that
//   implements methods to save, get and organize data via filenames, dirs, etc

// The filesystem must have access to a drive with geometry to access data.
// Geometry is page size and number of pages with methods to access. Also need
// way to chose which drive access is directed to. In the Cybiko case, the
// flash drives use a common SCI and a different chip enable. So the same
// routines can be used for access as long as different geometry is considered.

// The filesystem here is not like UNIX. No devices are mapped, etc. Every
// device is supported by its own API since the Cybiko has known fixed devices.
// External devices can have a support library written. (maybe used like DLL).

// The kernel has file systems and devices connected together.
// If a file system wants to do something, it has to get the device info from
// the kernel - via the DataSystem pointer (via File pointer)



/////////////////////////////////////   F I L E S Y S T E M
typedef struct
{
    // read, write, etc...
    char    id[FILESYSTEM_ID_MAX + 1];          // name/ID of file system
    StatusT (*init)(DataSystemT *);            // initialize the filesystem
    StatusT (*read)(int, int, int);        // read a data file
    StatusT (*flush)(int, int, int);       // flush a data file to device
    StatusT (*open)(LocationT *location);      // flush a data file to device
    StatusT (*write)(int, int, int);       // write a data file
    StatusT (*entrycreate)(int, int, int);     // create entry
    StatusT (*entrydelete)(int, int, int);     // delete entry
    StatusT (*custom)(int, int, int);          // custom command
} FileSystemT;




// support flash, RAM, etc...
/////////////////////////////////////   D E V I C E
typedef struct device
{
    // name or id of this device
    char    id[DEVICE_ID_MAX + 1];          // name/ID of the device

    // geometry of the device
    UINT16  pages;      // number of pages (aka blocks)
    UINT16  pagesize;   // how big are the pages
    UINT32  *pageused;  // points to array that keeps track of pages used
    void    *data;      // extra data to be appended if needed

    // read, write, etc...
    StatusT (*read)(PageT, size_t, size_t, char *);    // read
    StatusT (*readpage)(int, int, int);        // read whole page
    StatusT (*write)(int, int, int);           // write
    StatusT (*writepage)(int, int, int);       // write whole page
    StatusT (*takepage)(struct *device, PageT *page);  // get a free page
    StatusT (*givepage)(struct *device, PageT page);   // free a page
} DeviceT;


typedef char DriveLetterT;

typedef char *PathT;



/////////////////////////////////////   D A T A S Y S T E M
typedef struct
{
    DriveLetterT    token;          // identifier like A:  C:  etc...
    FileSystemT     *filesystem;    // filesystem overlayed on device

    DeviceT         *device;        // filesystem overlayed on this device
    PageT           start;          // page this DataSystem starts at
    PageT           pages;          // how many pages used by this drive

    // holds open pages on this data system - allocated dynamically
    PageT           *index;         // table of open pages

    // maybe put some kind of a cache here?
    void            *data;          // extra data for use

} DataSystemT;                      // could also be called a drive?





//
// structure for easy access - a path can be turned into a location in the
// DataSystem for programmatical use. All entries start on a page and belong
// to a directory (except the top dir). The structure of the entry itself
/// depends upon the file system.
typedef struct
{
    DataSystemT     *drive;     // what drive this entry is on - flash, ram?
    PageT           directory;  // page of directory found in
    PageT           page;       // page of entry
    PositionT       position;   // where is the pointer to the next byte to read

    // other things can be added for ease
//    size_t          size;       // size of the file
//    size_t          pagesize;   // how big are the pages
} EntryLocationT;


// The filesystem API -
// The filesystem will have directories and files. Directories can point to
// directories and files; files cannot point to anything and just end. The
// file system will use paths just like DOS or UNIX.
// Example path:
//    /0/stuff/file1  - file1 is a file, stuff is a dir, and 0 is the internal
//                      flash drive

// restriction: there will be a maximum of about 120 entries per directory
// also: only one filesystem supported - no reiserfs, ntfs, etc...
// A filesystem knows how things are linked together, what pages are, entries...


// DataSystem -
// Think of this as a drive in DOS: a: b: c:   etc...
// It is a filesystem overlayed on to a page device: DataFlash, Flash, RAM, etc


// Paths -
// All paths must be fully qualified. They must include the flash drive and
// any directories that lead to the file (this could change). The concept of a
// current directory should be kept in a shell/environment variable; the
// filesystem doesnt care about a current dir.

// Entry -
// An entry is either a directory or a file. The structure of the entry depends
// on the file system including the filename length, # of dir entries, etc..

// Directories -
// Directories can point to other directories or files (dirs contain entries).

// Files -
// A file contains data of any type. It can be opened for reading or writing.
// This API is only concerned that they can be created or deleted.
// See fileops.c for the API for working with Files.


// Drive API
// called once to initialize the data system - drives are discovered - dstab
// Devices must exist
StatusT DataSystemInit(void);


//////////////////////////////////////////////////////// Make this an app?
// Drive API
// Format the given drive - the whole thing gets formatted
StatusT DriveFormat(PathT path, FileSystemT *system);


// Entry API
// Create a new entry, directory or file depending on type
// There is a limit to how many entries can be added to a directory
StatusT EntryCreate(PathT path, EntryNameT name, EntryTypeT type);


// Entry API
// The entry will be deleted. If it is a directory, all connected dirs and files
//  will be deleted. Top directory cannot be deleted. Entries with busy files
//  cannot be deleted.
// // OR - cannot delete a directory with any entries in it
StatusT EntryDelete(PathT path);



/////////////////////////////////////////////////////////  NEEDED?

// Entry API
// Move the entry from one directory to another
StatusT EntryMove(PathT to, PathT from, EntryNameT *name);


// Entry API
// Copy the entry from one directory to another
StatusT EntryCopy(PathT to, PathT from, EntryNameT *name);



// end of header file
#endif
