#ifndef _GFX_H_
#define _GFX_H_
#include "types.h"
#include "rect.h"
#include "surface.h"



// four colors - 2 bits per pixel - fixed in the cybiko
// Use these for patterns, etc
#define GFX_COLOR_WHT   0x00
#define GFX_COLOR_LTG   0x55
#define GFX_COLOR_DKG   0xAA
#define GFX_COLOR_BLK   0xFF


#define GFX_MODE_PUT    0x00
#define GFX_MODE_OR     0x01


typedef unsigned char *GfxDisplayT;
typedef unsigned char GfxColorT;
typedef unsigned char GfxModeT;
typedef unsigned char GfxPatternT;

// Graphics point
typedef signed short GfxPT;

// Graphics Coordinate
typedef UINT GfxCT;

// The structure that is used to make calls with
typedef struct
{
    GfxDisplayT display;
    GfxColorT   color;
    GfxModeT    mode;
    GfxPatternT pattern;
} GfxContextT, *pGfxContextT;



// masks to help things go faster
// some are found in the monitor ROM
extern const unsigned char gfxleftmask[];
extern const unsigned char gfxrightmask[];
extern const unsigned char gfxcolormask[];
extern const unsigned char gfxcolors[];
extern const unsigned char gfxpixelmask[];
extern const unsigned char gfxpixelnotch[];




// A pixmap contains PIXEL map - for the cybiko it will always be defined as
//  a 4 gray map. Every 2 bits means ONE pixel. So width for a pixmap will
//  be twice as wide as a bitmap but the 'w' will be the same.
typedef struct
{
    UINT16  stride;         // how many bytes wide
    UINT16  width;          // active width in pixels
    UINT16  height;         // height = number of raster lines
//    unsigned char   transparent;    // the transparency color
//    unsigned char   flags;          // special information about the pixmap
    UCHAR   *image;         // the data for the pixmap
} PixelmapT;



///////////////////////////////////////////////////////////////////
// Points/Pixels and lines
///////////////////////////////////////////////////////////////////
void
GfxPixel(SurfaceT *surface, GfxCT x, GfxCT y);

// clears the pixel to zero
inline void
GfxPixelClr(SurfaceT *surface, GfxCT x, GfxCT y);

void
GfxLine(SurfaceT *surface, GfxPT x1, GfxPT y1, GfxPT x2, GfxPT y2);

void
GfxHLine(SurfaceT *surface, GfxPT x1, GfxPT y1, GfxPT x2);

void
GfxVLine(SurfaceT *surface, GfxPT x, GfxPT y1, GfxPT y2);
///////////////////////////////////////////////////////////////////




///////////////////////////////////////////////////////////////////
// Shapes
///////////////////////////////////////////////////////////////////
void
GfxBox(SurfaceT *surface, GfxPT x1, GfxPT y1, GfxPT x2, GfxPT y2);

void
GfxBoxX(SurfaceT *surface, RectT *rect);

void
GfxRect(SurfaceT *surface, GfxPT x1, GfxPT y1, GfxPT x2, GfxPT y2);

void
GfxRectX(SurfaceT *surface, RectT *rect);

void
GfxCircle(SurfaceT *surface, GfxPT xc, GfxPT yc, unsigned radius);


// NOT IMPLEMENTED YET
//void
//GfxCircleX(SurfaceT *surface, RectT *rect);
///////////////////////////////////////////////////////////////////




////////////////////////////////////////////////////////////////////
// Pixmaps - Pixelmap operations
///////////////////////////////////////////////////////////////////

void
GfxPixmapBlit(SurfaceT *surface, const PixelmapT *pixelmap, unsigned x, unsigned y);



/////////////////////////////////////////////////////////////////////////////
// Sprites - Sprites have a bitmap and position associated with the image.
///////////////////////////////////////////////////////////////////
// The bitmap indicates the used positions

typedef struct
{
    signed  x;              // the x position
    signed  y;              // the y position
    signed  xv;             // the x velocity
    signed  yv;             // the y velocity
    PixelmapT *pixelmap;    // what pixelmap to use
} SpriteT;



// draw the sprite to the screen
void
GfxSpriteDraw(GfxContextT *context, SpriteT *sprite);


// Move the sprite
void
GfxSpriteMove(SpriteT *sprite, GfxPT x, GfxPT y);


// Move the sprite and bounce of the bounding area
void GfxSpriteBounce(SpriteT *sprite, BoundsT *bounds);


// returns true if the sprites intersect
int
GfxSpriteCollide(SpriteT *sprite1, SpriteT *sprite2);


/////////////////////////////////////////////////////////////////////////////
////   F  O  N  T  S

// Fonts will be represented in storage as glyphs
// When a font is opened, it will be expanded into memory as a list of pixmaps

typedef enum
{
    FONT_IS_SCALEABLE     = 0x01,     // variable size - fixed width
    FONT_IS_PROPORTIONAL  = 0x02,     // fixed size, proportional width
    FONT_IS_FIXED         = 0x04,     // fixed size, fixed width
    FONT_IS_FAST          = 0x08,     // fixed size, fixed width byte aligned
} FontAttrT;


typedef enum
{
    FONT_TYPE_BITMAP,
    FONT_TYPE_SCALEABLE,
    FONT_TYPE_FAST,
} FontTypeT;


// data follows this header
typedef struct
{
    UINT8   width;
    UINT8   yoffset;
    UCHAR   data[];
} GlyphT;


// incomplete data type for font
//typedef struct font *FontT;
typedef struct font
{
    FontTypeT   type;
    FontAttrT   attributes;
    UCHAR       first;
    UCHAR       last;
    UCHAR       maxw;
    UCHAR       maxh;
    GlyphT      *glyph;
} *FontT;


void
GfxFont1(SurfaceT *surface, unsigned char data, RectT *rect);

typedef UINT32 FontIdT;

FontT
GfxFontOpen(FontIdT id);

UINT
GfxFontClose(FontT font);

UINT   // returns width of characters placed
// rect represents the size/spacing of chars - x is changed
GfxFontRender(SurfaceT *surface, FontT font, UCHAR *string, RectT *rect);

#endif

