// CBFS - Cybiko Reborn File System
// Internal use only - used by the DataSystem
// User should not call these directly




// At the open of a file, the file can be traversed to build an index
//



typedef char EntryTypeT;


// all CBFS names are the same
#define ENTRY_NAME_LEN 15
typedef char[ENTRY_NAME_LEN+1] EntryNameT;  // and zero terminated
typedef UINT16  EntrySizeT;    // entire size of the entry
typedef long    EntryDateT;    // date of the entry

// Every entry looks like this. Directory or File Entry
typedef struct
{
    PageT           next;   // index of the next page after this one - 0 = none
    EntrySizeT      size;   // bytes of data in data portion of entry
    EntryNameT      name;   // name of entry - zero filled if short
    EntryTypeT      type;   // top dir, dir, bcd, unix, dos, etc...
    EntryDateT      date;   // create date
} EntryHeaderT;



typedef struct
{
    EntryHeaderT    header;
    PageT           page[];
} DirPageT;


typedef struct
{
    EntryHeaderT    header;
    UINT8           data[];
} FilePageT;


typedef struct
{
    PageT           next;   // index of the following block of data - 0 = last
    UINT8           data[]; // the data in the rest of the page
} DataPageT;



#define FILESYSTEM_ID "CBFS"


// For the world to see:  // real declaration needs a jumptable? - no API do
FileSystemT cbfs =
{
    FILESYSTEM_NAME,
    CBFS_init,
    CBFS_fileseek,
    CBFS_filetell,
    CBFS_fileflush,
    CBFS_filewrite,
    CBFS_fileread,
    CBFS_entrycreate,
    CBFS_entrydelete,
    CBFS_entrysearch,
    CBFS_custom,
};



/////////////////////////////////////////////////

// look through directory to find if 'name' is there
// if it is found, the block is changed to the block containing name and ST_OK
// name can be for a directory or data entry, but the block must be a dir type
StatusT
CBFS_entrysearch(EntryLocationT *location, char *name)
{
    EntryHeaderT    header;
    StatusT         status;

    // if the block has the name, return true
    // the block we are starting at has a list of data to check
    // raw block read the entire block
    ASSERT(location != NULL);
    ASSERT(name != NULL);

    // read in the directory entry header
    status = location->drive.device->read(
            location->directory,
            0,
            sizeof(EntryHeaderT),
            &header);

    if (status != ST_OK) return status;

    // must be a directory that was read - cannot search data file
    if (header.type & CBFS_TYPE_DIR) == 0) return ST_NOTDIR;

    // now we have the header - is there any data in it?
    if (header.size != 0)
    {
        //  the data starts after the header - 100 entries per directory
        int     entries;
        PageT   entrypages[100];
        PageT   *searchindex;

        // read in the data in the dir entry all at once
        // overlay it on array of entry pages
        status = location->drive.device->read(
                location->directory,                // page to look in
                sizeof(EntryHeaderT),               // start byte
                header.size,                        // how many bytes
                entrypages);                        // where to put the data

        if (status != ST_OK) return status;

        // how many entries are in this block (could be more blocks)
        entries = header.size / sizeof(PageT);

        // helper for searcing the entries
        searchindex = entrypages;

        // Directory only has entries that will fit in one pagesize
        while (entries--)
        {
            EntryNameT  searchname;

            // read the name from the page
            status = location->drive.device->read(
                    *searchindex,                       // page
                    offsetof(EntryHeaderT, name),       // start of name
                    sizeof(EntryNameT),                 // the whole name
                    searchname);                        // name buffer

            if (status != ST_OK) return status;

            // is the name in the entry page?
            if (strcmp(name, searchname) == 0)
            {
                // found it
                location->page = *searchindex;
                return ST_OK;
            }

            // look in the next page
            searchindex++;
        }
    }

    return ST_NOTFOUND;
}


// Called from EntryDelete API call
// CyBorn File system CBFS - entry create
// filesystem specific checks, etc are made in here
StatusT
CBFS_entrydelete(EntryLocationT *location)
{
    StatusT         status;
    EntryHeaderT    header;


    // can only add an entry to a directory
    // use the index to check for the type of entry in the page
    // read in some of the page to be deleted

    // read in the directory entry header
    status = location->drive.device->read(
            location->directory,
            0,
            sizeof(EntryHeaderT),
            &header);

    if (status != ST_OK) return status;


    if (header.type & INDEX_DIR_CLASS)
    {
        // does the directory have any entries?
        // cannot delete a directory with entries!
        // read in the rest of the entry and search for any remaining entries
    }


    // Lock the filesystem - we are about to change it
    // CriticalEnter(FilesystemLock);

    // how many entries are there in this directory?
    DirEntryCount = header.size / sizeof(PageT);


    // load the data for the directory and search for the page entry in the
    // directory listing, remove the entry by adjusting the list and set
    // the size to one less entry
//    status = location->device.readpage(location->directory, buffer);
    if (status == ST_OK)
    {
        while (not done)
        {
            if (*DirectoryEntry == location->page)
            {
                // found it -  remove it
                // copy the last one here over the one to delete
                DirectoryEntry = header.data[numberofentriesindir];
                header.size -= sizeof(PageT);

                // put the header into the empty block
                location->drive->device.writepage(page, header, 0x00, sizeof(EntryHeaderT));
            }
        }
    }

    // CriticalExit(FilesystemLock);

    return status;
}


// Called from EntryCreate API call
// CyBorn File system CBFS - entry create
// filesystem specific checks, etc are made in here
StatusT
CBFS_entrycreate(EntryLocationT *location, EntryNameT name, EntryTypeT type)
{

    // can only add an entry to a directory
    // use the index to check for the type of entry in the page
    if (location->drive->index[page] & INDEX_DIR_CLASS == 0) return ST_ILLEGAL;

    // any room left in the directory for another entry?
    if (location->header.size ==
        location->drive->device.pagesize - sizeof(EntryHeaderT)) return ST_FULL;


    // cannot add a name that is already in the directory
    // look for an entry name in the directory page
    // if the name is found, then this is a duplicate name
    // - Read the directory page into buffer
    location->drive->device.select(TRUE);
    location->drive->device.readpage(location->directory, buffer);
    location->drive->device.select(FALSE);

    // any room left in the directory for another entry?
    if ((EntryHeaderT *)buffer->size ==
        location->drive->device.pagesize - sizeof(EntryHeaderT)) return ST_FULL;


    // how many entries are there in this directory?
    DirEntryCount = ENTRY_SIZE(buffer) / sizeof(PageT);
    EntryNumber = (PageT *)(buffer + sizeof(EntryHeaderT));
    while (DirEntryCount--)
    {
        char    entryname[20];

        // check the entry name
        // Read the name from the page
        // TODO - make THREAD safe (put inside the .select?1
        location.drive->device.select(TRUE);
        location.drive->device.read(*EntryNumber++, entryname, 20);
        location.drive->device.select(FALSE);

        if (strcmp(entryname+NAMEOFFSET, name) == 0) return ST_DUPLICATE;
    }


    // Lock the filesystem - we are about to change it
    // CriticalEnter(FilesystemLock);

    // get a free block
    status = GetFreeBlock(&block);
    if (status == ST_OK)
    {
        header.next = 0;
        header.size = 0;
        memset(header.name, 0, sizeof(EntryNameT));     // zero filled
        strcpy(header.name, name);
        header.type = type;

        // add the entry to the directory - just a page as far as the dir cares
        EntryAdd(&location, block);

        // put the header into the empty block
        location->drive->device.write(block, header, 0x00, sizeof(EntryHeaderT));

        // mark that the block is used and not free. It does not have a next
        location->drive->index[block] = INDEX_PAGE_USED;
    }

    // CriticalExit(FilesystemLock);

    return status;
}



// Called from read API call
// CyBorn File system CBFS - read
// filesystem specific checks, etc are made in here
StatusT
CBFS_fileread(EntryLocationT *location, size_t *size, char *buffer)
{
    long            from;
    PageT           next;
    EntryHeaderT    header;
    size_t          sizeleft;
    StatusT         status;
    size_t          sizetoread;
    size_t          limit;
    short           fullpage;

    // parameter check
    ASSERT(location != NULL);
    ASSERT(size != NULL);
    ASSERT(buffer != NULL);

    // silly, but if size is zero, we are finished
    if (*size == 0) return ST_OK;

    // Read the entry header from the first page of the entry
    status = location->drive->device.read(location->page, 0, sizeof(header), &header);
    if (status != ST_OK) return status;

    // cant read more than what is left in the file - set a limit
    limit = header.size - location->position.logical;

    // are we at the end of the file?
    if (limit == 0) return ST_EOF;

    // may have requested a size that is bigger than what is left
    sizeleft = (*size > limit) ? limit : *size;

    // now set the size to zero and add only when real bytes are read
    *size = 0;


    // read from the first page of data

    // not reading to the end of the first page? eg only 20 bytes when 30 remain
    if (sizeleft + location->position.offset < location->pagesize)
    {
        // dont have to read the whole page - or another
        // read the data - as mush as possible into the buffer
        status = location->drive->device.read(
                location->position.page,
                location->position.offset,
                buffer,
                sizeleft);
        if (status != ST_OK) return status;

        location->position.offset += sizeleft;
        location->position.logical += sizeleft;

        *size = sizeleft;

        return ST_OK;
    }

    // read to the end of the page
    sizetoread = location->drive.pagesize - location->position.offset;
    status = location->drive->device.read(
            location->position.page,
            location->position.offset,
            buffer,
            sizetoread);
    if (status != ST_OK) return status;

    // add the bytes that were read to advance the position
    location->position.logical += sizetoread;

    // set size
    *size = sizetoread;

    // have we read to the end of the file?
    if (header.size == location->position.logical) return ST_OK;

    // get the next page
    status = location->drive->device.read(
            location->position.page,
            0,
            sizeof(next),
            &next);
    if (status != ST_OK) return status;

    // need to point to next data page
    location->posiition.page = next

    // bump the pointers
    buffer += sizetoread;
    sizeleft -= sizetoread;

    // start over at the start of data in a page
    location->position.offset = sizeof(DataPageT);

    ///////
    // read the middle - if there is more than a page left - read whole pages
    fullpage = location->drive->device.pagesize - sizeof(DataPageT);
    while (sizeleft >= fullpage)
    {
        // read the data - full page of data
        status = location->drive->device.read(
                location->position.page,
                location->position.offset,
                fullpage,
                buffer
                );
        if (status != ST_OK) return status;

        // add the bytes that were read to advance the position
        location->position.logical += fullpage;
        *size += fullpage;

        // have we read to the end of the file?
        if (header.size == location->position.logical) return ST_OK;

        // read the header to get the next page
        status = location->drive->device.read(
                location->position.page,
                0,
                sizeof(next),
                &next);
        if (status != ST_OK) return status;

        // save the new position
        location->position.page = next;

        // fixup the number left to retrieve
        buffer += fullpage;
        sizeleft -= fullpage;
    }

    // read the last part - less than a page left - wont have to bump the page
    // and will not end reading on a page boundry
    if (sizeleft)
    {
        status = location->drive->device.read(
                location->page,
                location->position.offset,
                sizeleft,
                buffer);
        if (status != ST_OK) return status;

        location->position.offset += sizeleft;
        location->position.logical += sizeleft;
        *size += sizeleft;
    }

    return ST_OK;
}






// Called from FileSeek API call
// CyBorn File system CBFS - fileseek
// filesystem specific checks, etc are made in here
StatusT
CBFS_fileseek(EntryLocationT *location, size_t offset, FileSeekT type)
{
    StatusT         status;
    EntryHeaderT    header;
    long            LogicalPosition;

    // the given input must be converted to a page and offset to fill in the
    /// position

    // error check
    ASSERT(location != NULL);

    // read the header to get the size
    // cache this?
    status = location->drive->device.read(location->page, 0, sizeof(header), &header);
    if (status != ST_OK) return status;

    switch (type)
    {
        case FILE_SEEK_END:
            // the relative offset from the end
            LogicalPosition = header.size + offset;
            break;

        case FILE_SEEK_CUR:
            // the relative offset from the current position
            LogicalPosition = location->position.logical + offset;
            break;

        default:
        case FILE_SEEK_SET:
            LogicalPosition = offset;
            break;
    }

    // error check
    if (LogicalPosition > header.size) return ST_WRONG;
    if (LogicalPosition < 0) return ST_WRONG;

    // convert the logical position into a physical position
    // start at the first page

    if (LogicalPosition < location->drive.pagesize - sizeof(EntryHeaderT))
    {
        // its inside the start page
        location->position.page   = location->page;
        location->position.offset = sizeof(EntryHeaderT) + LogicalPosition;
    }
    else  // not in the first page
    {
        short   pages;
        PageT   next;

        pages = (LogicalPosition - location->drive.pagesize + sizeof(EntryHeaderT))
                / (location->drive->device.pagesize - sizeof(DataPageT)) - 1;

        next = header.next;

        // this can be time consuming if we start at the beginning and have
        // to go far to the end
        while (pages--)
        {
            status = location->drive->device.read(next, 0, sizeof(next), &next);
            if (status != ST_OK) return status;
        }

        location->position.page = next;
        location->position.offset =
            (LogicalPosition - location->drive.pagesize + sizeof(EntryHeaderT))
            % (location->drive.pagesize - sizeof(DataPageT));

    }

    // set the logical position
    location->position.logical = LogicalPosition;

    return ST_OK;
}



// fill in the position that only the filesystem would know about
// The calling function should have done the heavy lifting

StatusT
CBFS_fileopen(EntryLocationT *location)
{
    // error check
    ASSERT(location != NULL);

    // fill in the stuff that is important to the filesystem
    location->position.page = 0;
    location->position.offset = sizeof(DataPageT);
    location->position.logical = 0;

    return ST_OK;
}



// Where is the next byte read from
StatusT
CBFS_filetell(EntryLocationT *location, size_t *size)
{
    // error check
    ASSERT(location != NULL);
    ASSERT(size != NULL);

    // return the logical position
    *size = location->position.logical;

    return ST_OK;
}



StatusT
CBFS_fileflush(EntryLocationT *location)
{
    StatusT status;


    // error check
    ASSERT(location != NULL);
    ASSERT(location->drive.data);

    // time to write the page
    if (something->flag & WRITE_BUFFER_DIRTY)
    {
        status = location->drive->device.writepage(
                location->position.page,
                0,
                location->drive.pagesize,
                location->drive.data);

        something->flag &= ~WRITE_BUFFER_DIRTY;
    }

    return ST_OK;
}









