// All Data flash commands and operations are the same
// Some Atmel data flash devices have different page sizes - so reading a
/// page may return more data on one device than another

#define ATMEL_CMD_READMAIN      0x52
#define ATMEL_CMD_STATUS        0x57
#define ATMEL_CMD_READPAGE      0x52
#define ATMEL_CMD_WRITEPAGE     0x82    // uses buffer 1


typedef struct
{
    UINT8   mra;
    UINT8   sar[3];
    UINT8   mrb;
    UINT8   dar[3];
    UINT16  cra;    // make sure the byte ordering is correct for the part
    UINT16  crb;
} DtcRegistersT;



typedef union
{
    UINT32  allofit;
    UINT8   parts[4];
} PageByteT;


typedef struct
{
    char command;
    PageByteT pagebyte;
    char dummy[4];
} AtmelReadCommandT;

// The vectors for DTC are set in the monitor rom to:
// #define DTC_SERIAL1_RX_VECT 0x04AA
// #define DTC_SERIAL1_TX_VECT 0x04AC
// they contain FBD0 and cannot be changed since in ROM


// Tx and Rx use same DTC in ram - location in RAM is preset by bootloader ROM
/// the write can chain to the next for command then data - for write
#define DTC_SERIAL1_TX ((DtcRegistersT *)0xFFFBD0)
#define DTC_SERIAL1_RX ((DtcRegistersT *)0xFFFBD0)
#define DTC_SERIAL1_NEXT ((DtcRegistersT *)0xFFFBDC)



static long
pagetoaddress(DeviceT *device, PageT page)
{
    // account for different page sizes
    switch (device->type)
    {
        default:
        case 0: return (long)page << 9;
        case 1: return (long)page << 10;
        case 2: return (long)page << 11;
        case 3: return (long)page << 12;
    }
}

#define DTC_SAR_FIXED   0x00
#define DTC_DAR_FIXED   0x00

#define DTC_SAR_INC     0x80
#define DTC_DAR_INC     0x20

#define DTC_SAR_DEC     0xC0
#define DTC_DAR_DEC     0x30

#define DTC_MODE_NORMAL 0x00
#define DTC_MODE_REPEAT 0x04
#define DTC_MODE_BLOCK  0x08

#define DTC_SELECT_DST  0x00
#define DTC_SELECT_SRC  0x02

#define DTC_SIZE_BYTE   0x00
#define DTC_SIZE_WORD   0x01

#define DTC_TRANS_END   0x00
#define DTC_TRANS_CHAIN 0x80

#define DTC_INT_DISABLE 0x00
#define DTC_INT_ENABLE  0x40



void
_device_dataflash_read(DeviceT *device, PageT page, UINT8 *buffer, size_t from, size_t size)
{
    PageByteT pagebyte;


    ASSERT(device != NULL);
    ASSERT(buffer != NULL);

    // account for different page sizes
    pagebyte.allofit = pagetoaddress(device, page);
    pagebyte.allofit |= from;

    // move the page at current to page buffer
    // need to have exclusive access to the serial port
// SignalWait(Serial_1);

    // are all other operations on this chip finished? check and wait if needed
    AtmelWait();

    // The chip, serial port are ours and ready - send the command to read
    // The command will be sent to the chip
    // The data will then be read back from the chip into our buffer
    DTC_SERIAL1_TX->mra =
            DTC_SAR_INC |       // the source pointer increments
            DTC_DAR_FIXED |     // the dest is fixed
            DTC_MODE_NORMAL |   // normal mode
            DTC_SIZE_BYTE;      // byte transfers

    DTC_SERIAL1_TX->mrb =
            DTC_TRANS_CHAIN |   // chain to the next transfer
            DTC_INT_DISABLE;    // dont do interrupt - DTC clears it

    // set the source and destination
    // the serial port needs to be set up
    DTC_SERIAL1_TX->sar = command;      // the command that we have buffered
    DTC_SERIAL1_TX->dar = &SCI1.TDR;    // send the command to the transmit buf


    DTC_SERIAL1_TX->cra = sizeof(command);
    // DTC_SERIAL1_TX->crb =


    // Read data back from the chip into our buffer
    DTC_SERIAL1_TX->mra =
            DTC_SAR_FIXED |   // the source is fixed at the Receive buffer
            DTC_DAR_INC |     // the dest incremensts
            DTC_MODE_NORMAL |   // normal mode
            DTC_SIZE_BYTE;      // byte transfers

    DTC_SERIAL1_TX->mrb =
            DTC_TRANS_END |     // last one
            DTC_INT_DISABLE;    // dont do interrupt - DTC clears it

    // set the source and destination
    // the serial port needs to be set up
    DTC_SERIAL1_TX->sar = &SCI1.RDR;    //
    DTC_SERIAL1_TX->dar = buffer;    // send the command to the transmit buf

    DTC_SERIAL1_TX->cra = size;
    // DTC_SERIAL1_TX->crb =

    // TRIGGER the whole thing
    // how do we know when we are done?

// Signal(Serial_1);

// CriticalExit(FlashAccess);
}



// Assume owns serial port and chip select for desired atmel is asserted
// All dataflash access goes through serial port 1 at same speed
static void inline
AtmelWait(void)
{
    // wait until the transmit data register empty flag is asserted
    // TDRE = Transmit data register empty - wait 'til true
    // Warning - this will spin here until the bit is set - conditions must be
    /// setup so that this can happen (serial system is setup)
    while (!SCI1.SSR.BIT.TDRE) ;

    // put the command in the Transmit Data Register
    SCI1.TDR = ATMEL_CMD_STATUS;


    // Now we want data in from the serial line - what will make the clock?
    SCI1.SSR.BIT.ORER = 0;      // clear error

    // Set the receive enable
    SCI1.SCR.BIT.RE = 1;

    // receive the status byte
    do
    {
        // any error?
        if (SCI1.SSR.BIT.ORER)
        {
            // got an overrun error
            SCI1.SSR.BIT.ORER = 0;      // clear error
        }

    } while (!SCI1.SSR.BIT.RDRF);

    // got status - no more receiving
    SCI1.SCR.BIT.RE = 0;


#if 0
    syncserialout(DATAFLASH_CMD_STATUS);

    // now clock only one data line in
    // bit twiddle the clock once
    somecommand();

    // wait until the bit goes high and is ready
    // TODO - make threaded or use sleeps, use interrupt to give sem
//    while (portx.bx) sleep(1);
    while (portx.bx) ;
#endif
}



// write the whole page
static void
_device_dataflash_writepage(DeviceT *device, PageT page, UINT8 *buffer)
{
    UINT16 size;

    PageByteT pagebyte;


    ASSERT(device != NULL);
    ASSERT(buffer != NULL);


    pagebyte.allofit = pagetoaddress(device, page);

    size = device->pagesize;

    // move the page at current to page buffer
// CriticalEnter(FlashAccess);  // system wide - all flash access is common
    atmelwait();  // TODO SignalWait(device->ready);
    syncserialout(ATMEL_CMD_READMAIN);
    syncserialout(pagebyte.parts[2]);
    syncserialout(pagebyte.parts[1]);
    syncserialout(pagebyte.parts[0]);
    while (size--)
    {
        syncserialout(*buffer++);
    }
// CriticalExit(FlashAccess);
}


typedef struct
{
    char command;
    PageByteT pagebyte;
    char dummy[4];
} AtmelReadCommandT;




// Transmit interrupt - happens when the transmit is done
void interrupt
external_flash_ready(void)
{
    // called at the end of the transfer
    // chip select deassert
    PortX.b = 1;


//    Signal(Serial1);
}


// Transmit interrupt - happens when the transmit is done
void interrupt
internal_flash_ready(void)
{
    // called at the end of the transfer
    // chip select deassert
    PortX.b = 1;

//    Signal(Serial1);
}











// MACRO this?

void
_device_DTC_readpage(DeviceT *device, PageT page, char *buffer)
{
    _device_DTC_read(device, page, buffer, 0, device->pagesize);
}



//  The device select must be asserted for this

void
_device_DTC_read(DeviceT *device, PageT page, char *buffer, size_t from, size_t size)
{
    AtmelReadCommandT command;


    ASSERT(device != NULL);
    ASSERT(buffer != NULL);


    // device->select(assert);    -   portX.b = 0;
    // device->select(deassert);  -   portX.b = 1;

    command.pagebyte.allofit = pagetoaddress(device, page) | from;

    // move the page at current to page buffer
// CriticalEnter(FlashAccess);
    atmelwait();

    // the DTC will chain - the first will be a command to write
    // The next will be a DTC that will read the data back to the buffer

    //  VERY IMPORTANT - reading the size less than 2 is not well supported
    if (size < 2) size = 2;
    // the buffer MUST SUPPORT the extra size
    // TODO change this to have a local long capture at least 4 bytes
    /// then copy the 1 byte to the buffer if size is < 4

    // set up DTC command block
    DTC_SERIAL1_TX->sar = &command;   // Source Address of command
    DTC_SERIAL1_TX->dar = SERIAL2_TX;   // Destination - TX register for Serial
    DTC_SERIAL1_TX->mra = 0xXX;   // Mode register A - inc src
    DTC_SERIAL1_TX->mrb = 0x80;   // Mode register B
    DTC_SERIAL1_TX->cra = sizeof(AtmelReadCommandT);    // how many bytes send

    // vector is already preset in the monitor ROM

    // enable the transfer - write the command out
    // set the tx as needed

    // set up DTC receive block - chained
    DTC_SERIAL1_NEXT->sar = SERIAL2_RX;   // source address
    DTC_SERIAL1_NEXT->dar = buffer;      // buffer for receiving data
    DTC_SERIAL1_NEXT->mra = 0xXX;   // Mode register A - inc dest
    DTC_SERIAL1_NEXT->mrb = 0x80;   // Mode register B
    DTC_SERIAL1_NEXT->cra = size;         // at least 2!!!!!!!!!!!!!!

    // enable the transfer - write the command out
    // set the rx to begin the transfer from the memory to the buffer

    // when finished, the cleanup interrupt will set
    /// flash_is_ready to true

    // Now, make sure that the DTC is finished
    while (flash_is_ready == 0) ;

    // alternate:
    // SignalWait(device->ready);   // given by the interrupt when done


// CriticalExit(FlashAccess);


}











