

#include "fileops.h"
#include "malloc.h"


// known devices - the external device may have a different size and geometry
#define FILE_DEVICE_DATAFLASH_INTERNAL 1    // onboard serial flash
#define FILE_DEVICE_DATAFLASH_EXTERNAL 2    // dataflash on the data cart
#define FILE_DEVICE_RAMDISK_INTERNAL   3    // Internal memory
#define FILE_DEVICE_RAMDISK_EXTERNAL   4    // Data cartridge RAM
#define FILE_DEVICE_OSFLASH            5    // The OS flash extra


#define FILE_DEVICE_NAME_LEN 3
#define FILE_DEVICE_NAME_DLM ':'

#define FILE_DEVICE_DEFAULT FILE_DEVICE_DATAFLASH_INTERNAL


// how to distinguish internal and external flash/ram devices
// there may be a discovery phase by the OS and it could find different device


/////////////////////////////////////   F I L E
typedef struct file
{
    struct file     *next;      // use for linked list
    //struct file     *prev;
    EntryLocationT  location;   // location of this flash file
    int             mode;
} *FileT;




// Pointer to the list of all open files
FileT FileList;

// later get it from the kernel
//extern FileT FileList;



// in the Cybiko Classic, there can be more than one file system
DeviceT DataFlashInternal;
DeviceT DataFlashExternal;
DeviceT RamInternal;
DeviceT RamExternal;
DeviceT FlashInternal;
// others... could have an expansion card with more than one device...



// All of the data for the file needs to be filled in.
// A pointer is returned for the file
// Check and qualify then call the right
StatusT
FileOpen(FileT *file, FilePathT path, FileFlagsT flags)
{
    StatusT         status;
    FileT           openfile;
    EntryLocationT  location;

    // error check
    if (path == NULL) return ST_NULL;

    // locate the file to open
    // need a full path to the file to be opened
    status = LocationFromPath(path, &location);
    if (status != ST_OK) return status;

    // only one write access at a time so check if already open for write
    if (flags & FILE_MODE_WRITE)
    {
        // is the file open as write? Start at the front and search
        openfile = FileList;

        // go through the entire list of files to see if it is open as write
        while (openfile)
        {
            // start page has been found in same device and has open for write
            if ((openfile->start == location->page) &&
                        (openfile->drive == location->drive) &&
                        (openfile->mode & FILE_MODE_WRITE))
            {
                // file is already open as write
                return ST_WRITE;
            }

            // look at next open file
            openfile = openfile->next;
        }
    }

    // get the memory for the entry and location
    openfile = (FileT)malloc(sizeof(struct file));
    if (openfile == NULL) return ST_NOMEMORY;

    // have the file system open and init the file
    openfile->location->drive->filesystem.open(&location);

    // copy in the location
    // memcpy (openfile->location, location, sizeof(location));
    openfile->location = location;

    // connect the file into the OS database
    // MUTEX : CriticalEnter(FileListMutex);
    openfile->next = FileList;
    if (FileList)
    {
        FileList->next = openfile;
    }
    else
    {
        FileList = openfile;
    }
    // MUTEX : CriticalExit(FileListMutex);

    *file = openfile;

    return status;

}



// Adjust the position
StatusT
FileTell(FileT file, size_t *position)
{
    StatusT     status;


    if (file == NULL)       return ST_NULL;
    if (position == NULL)   return ST_NULL;

    return file->location.drive->filesystem.tell(&file->location, position);
}



// remove the entry from the directory
Status
RemoveEntry(EntryLocationT *location)
{
    // go to drive and find the 'entry' to delete in the directory

    // fix the index in the drive to show that the blocks are not used anymore
    // lock the list
    // do the stuff
    // unlock the list

    return ST_OK;
}




// This is the API for removing a file
StatusT
FileDelete(PathT *path)
{
    StatusT         status;
    EntryLocationT  location;
    FileT           file;

    // where is the entry?
    status = LocationFromPath(path, &location);
    if (status != ST_OK) return status;

    // is the file open? May not be deleted
    file = FileList;

    // go through the entire list of files to see if it is open
    while (file)
    {
        // start page has been found in same device
        if ((file->start == location->page) && (file->drive == location->drive))
        {
            // file is already open - cannot delete it
            return ST_OPEN;
        }
    }

    // ok to delete the file - use the directory and remove it
    // > open file for write access
    // > delete the entry from the directory
    // > update free blocks someplace
    // > close the file and remove from the kernel entry
    status = file->location.drive->filesystem.entrydelete(&location);

    return status;
}



// move the position somewhere - need to qualify?
StatusT
FileSeek(FileT file, size_t offset, FileSeekT type)
{
    PageT page;

    // simple error checking
    if (file == NULL) return ST_NULL;

    return file->location->filesystem.seek(&file->location, offset, type);
}



// not an error to flush a read only file - but does nothing
StatusT
FileFlush(FileT file)
{
    // simple error checking
    if (file == NULL) return ST_NULL;

    if (file->mode & FILE_MODE_WRITE)
    {
        // write the entire cached page to the flash device
        file->location->drive->filesystem.flush(file);
    }

    return ST_OK;
}



// could use magic numbers to confirm a real file pointer
StatusT
FileClose(FileT file)
{
    // simple error checking
    if (file == NULL) return ST_NULL;
    if (FileList == NULL) return ST_FILE;

    // where is the file? At the start?
// MUTEX : CriticalEnter(FlashFileListMutex);
    if (FileList == file)
    {
        FileList = file->next;
    }
    else
    {
        FileT   closefile = FileList;

        // must find it!!!
        for (;;)
        {
            if (closefile->next == 0)
            {
                // didnt find the file
                return ST_FILE;
            }

            // are we pointing to the file?
            if (closefile->next == file)
            {
                closefile->next = file->next;
                break;
            }

            // go to the next one
            closefile = closefile->next;
        }
    }
// MUTEX : CriticalExit(FlashFileListMutex);

    if (file->mode & FLASH_MODE_WRITE)
    {
        // we are guaranteed to be the only file with write priv
        // flush it
        //file->location->drive->filesystem.flush(file);
        FlashFlush(file);
    }



    // free the memory used
    //free(file->cache);
    free(file);

}




//
StatusT
FileRead(FileT file, UINT8 *buffer, size_t *size)
{
    // is this file handle any good?
    // could also qualify it by checking if it is in the OS list of files
    if (file == NULL) return ST_NULL;
    if (buffer == NULL) return ST_NULL;
    if (size == NULL) return ST_NULL;

    // call the filesystem read proceedure with params
    return file->drive->filesystem.read(&file->location, size, buffer);
}



