;
;    Cyborn OS - Copyright (C) 2004 Steve Gotthardt - Gotthardt Designs
;
;    This file is part of the Cyborn distribution.
;
;    Cyborn is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    Cyborn is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with Cyborn; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;    A special exception to the GPL can be applied should you wish to distribute
;    a combined work that includes Cyborn, without being obliged to provide
;    the source code for any proprietary components.
;

;   Low-level driver for the keypad of the cybiko classic
;   Each address line is driven low and the data is inspected to see if it is
;   low - meaning that a key is pressed and shorting the address line to data.
;
;   An OS should call this and also keep track of Shift and Fn keys. It should
;   also translate these numbers into ASCII. Could also do sticky shift, etc.

;   Re-entrant: yes - thread safe
;

;   Note
;   The Esc is no longer in the matrix for V2. It was at A1x1 for the V1 classic.
;   It is now mapped to port1.3

;   Note
;   The shift key cannot be held down with other keys in its data line
;   For example: shift + [ does not give the expected: {




;   Here is the data line and address line matrix - A0 doesn't matter
;
;
;                ----- D A T A L I N E -----
;             7    6    5    4    3    2    1    0 : bit position
;
;   A1      Shift  `    A    Q  left  del (esc) F7

;   A2       Fn    Z    S    W    2   ins   up  F6

;   A3      Help   X    D    E    3  space  F3  F5

;   A4       [     C    F    R    4   tab   1   F4

;   A5       ]     V    G    T    5  select dn  right

;   A6       \     B    H    Y    6  enter  ;   F2

;   A7             N    J    U    7  BkSp   /   F1

;   A8             M    K    I    8    0    .    -

;   A9             ,    L    O    P    9    =    '

    .h8300s
    .text

    .global __rawkey        ; returns the key processed in raw form
    .global __esckey        ; returns non-zero in r0 if esc key is pressed
    .global __keypressed    ; returns non-zero if there is a key pressed
    .global __keycheck      ; returns non-zero if requested key is pressed


; reads the whole pad at once looking for any switch closure
;   returns: r0 is non-zero if a key is pressed (except esc which is on a port)
__keypressed:
    mov     #0xEFFC00,  er0     ; turn on all address lines at once
    mov     @er0,       r0      ; must? read word at a time to get the switches
    not     r0                  ; read value will be 0xFF if no switch pressed
    rts                         ; so use 'not' to make it a zero



;   r0 returns keypress - 0 means NO key pressed
;   if you know what key you want try the _keycheck function
;   no input parameter
__rawkey:
    ; do the first two lines and ignore the modifiers: shift and fn
    sub     r0,         r0      ; clear working reg
    mov     #0xEFFFFD,  er2     ; Address bit 1 = 0 - first row

    mov     @er2,       r1      ; load up the key scan for this address line
    or      #0x80,      r1l     ; ignore that the shift might be pushed
    not                 r1l
    bne     foundline

    add     #8,         r0l
    rotl                r2      ; rotate the bottom half
    mov     @er2,       r1
    or      #0x80,      r1l     ; ignore that the 'fn' might be pushed
    not                 r1l
    bne     foundline

    add     #8,         r0l

nextline:
    rotl                r2      ; rotate address - only the lower half
    mov     @er2,       r1      ; must read a word at a time
    not                 r1l     ; 0xFF means no key is pressed in this line
    bne     foundline

    add     #8,         r0l     ; account for next row
    cmp     #72,        r0l     ; did we check all of the lines?
    bne     nextline

    sub     r0,         r0      ; clear return value - no key found
    rts

    ;found the address line - make the return value
    ;r1h has dataline pulled low and r0l has offset for which address line
foundline:
nextzero:
    inc                 r0l      ; start at bit position 0
    ;now look for the 1 in the byte
    shar                r1l
    bcc                 nextzero ;still looking for a one

retkey:
    rts




; returns: r0 = non-zero if the esc key is pressed
__esckey:
    mov     @0x50:8,    r0l     ; Read port 1
    not                 r0l
    and     #0x08       r0      ; mask off only the esc switch
    rts




;   the list of addresses for reading rows of keys
addlines:
    .long    0x0EFFFFD  ; bit 1 -  1111 1101
    .long    0x0EFFFFB  ; bit 2 -  1111 1011
    .long    0x0EFFFF7  ;          1111 0111
    .long    0x0EFFFEF  ;          1110 1111
    .long    0x0EFFFDF  ;          1101 1111
    .long    0x0EFFFBF  ;          1011 1111
    .long    0x0EFFF7F  ;          0111 1111
    .long    0x0EFFEFF  ;     1110 1111 1111
    .long    0x0EFFDFF  ;     1101 1111 1111



;   r0 = rawkey to check for - do NOT use ASCII here - use lowlevel.h value
;   returns non-zero in r0 if the key is pressed
__keycheck:
    dec     #1,         r0
    mov     r0,         r1
    shlr    r1                      ; shift to a multiple of 4
    and     #0x3C,      er1         ; mask only the needed bits
    mov     @(addlines,er1), er1    ; get the correct address
    mov     @er1,       r1l         ; read the keypad
    not     r1l

    and     #7,         er0         ; mask the good bits
    mov     @(_bitposition,er0), r0l ; get the mask
    and     r1l,        r0l

    rts




