
#include <string.h>
#include <stdlib.h>

#include "filesys.h"



// Functions that need to be exported for a RAM device that can be
//// used for a RAM data system.
DeviceFunctionsT ramfunctions =
{
    read,
    writepage,
    takepage,
    givepage,
};



typedef struct
{
    char *usedpages;
    char *RamBase;
    PageT pages;
    short pagesize;
} RamDriveSpecsT;




static StatusT
RamDeviceCreate(DeviceT *device, void *specs)
{
    // get space for extra RAM drive info
    device->pageused = (char *)calloc((RamDriveSpecsT *)device->pages / 32 + 1);
    if (device->data == 0) return ST_NOMEMORY;

    // save the specs
    device->pagesize = (RamDriveSpecsT *)specs->pagesize;
    device->pages = (RamDriveSpecsT *)specs->pages;

    // point data to start of data after the header
    (char *)device->data = (RamDriveSpecsT*)specs->RamBase;

    // copy in the function pointers for the device
    device->read = read;
    device->writepage = writepage;
    device->takepage = takepage;
    device->givepage = givepage;


    return ST_OK;
}



static StatusT
RamDeviceDestroy(DeviceT *device)
{
    // release the memory that was allocated during 'create'
    if (device->pageused)
    {
        free(device->pageused);
    }

    return ST_OK;
}




static StatusT
writepage(DeviceT *device, PageT page, char *data, size_t offset, size_t size)
{
    // write to RAM should be easy
    memcpy((char *)device->data + page * device->pagesize + offset, data, size);

    return ST_OK;
}




static StatusT
read(DeviceT *device, PageT page, size_t offset, char *data, size_t size)
{
    // just copy memory into buffer
    memcpy(data, (char *)device->data + page * device->pagesize + offset, size);

    return ST_OK;
}



static StatusT
takepage(DeviceT *device, PageT *page)
{
    int i, j, k;

    // find a free page
    for (i = 0; i < PAGE_BLOCK_SIZE/32; i++)
    {
        register unsigned long j;

        // skip if all pages are used in this page used record
        if (device->usedpages[i] == 0xFFFF) continue;

        k = 0;
        j = 0x8000;

        for (;;)
        {
            if (device->usedpages[i] & j == 0)
            {
                device->usedpages[i] |= j;
                *page = k + i * 32;
                return ST_OK;
            }

            // shift j and increment k
            j >>= 1;
            k++;
        }
    }

    // no open page
    return ST_FULL;
}



static StatusT
givepage(DeviceT *device, PageT page)
{
    if (device == 0) return ST_NULL;

    // release the used page
    device->usedpages[page/32] &= mask[page & 0x1F];

    return ST_OK;
}





