// Note: there is something wrong with binutils and using --relax for ld. Using
// the new binutils: 2.16 and the elf format, I dont get errors, but it doesn't
// seem to do any optimizations of :32 bit access in upper mem to :8

// There are no hardware handshake connections from the cybiko to the PC. Only
// ground, tx and rx. Should implement SW handshake as an option.


#include <stddef.h>
#include <stdlib.h>
#include <types.h>
#include "status.h"
#include "2245s.h"
#include "rs232.h"


// jump vectors in RAM
// The cybiko uses SCI2 for RS-232
#define ERI2_VECTOR *((UINT32 *)0xFFECAC)    // receive error
#define RXI2_VECTOR *((UINT32 *)0xFFECB0)    // rx data buffer full
#define TXI2_VECTOR *((UINT32 *)0xFFECB4)    // tx data buffer empty
#define TEI2_VECTOR *((UINT32 *)0xFFECB8)    // tx end


#define ASCII_CTRL_S    (19)      // default XOFF
#define ASCII_CTRL_Q    (17)      // default XON


#define TX_XOFF     (2)    // received an XOFF and should stop transmitting data
#define TX_XNOT     (1)
#define TX_XON      (0)

#define RX_XOFF     (2)    // sent an XOFF because we can receive no more data
#define RX_XNOT     (1)
#define RX_XON      (0)


#define BUFFER_MARGIN       (12)
#define QUEUE_SIZE_MIN      (64)


// where did I put the crystal frequency? Use it to get the 345600
// 345600 = 11059200 / 32
#define BAUD_REG(baud) ((345600 / (baud)) - 1)



// Transmit Enable and Disable
#define TXI_DISABLE()   SCI2.SCR.BIT.TIE=0
#define TXI_ENABLE()    SCI2.SCR.BIT.TIE=1

// Receive Enable and Disable
#define RXI_DISABLE()   SCI2.SCR.BIT.RIE=0
#define RXI_ENABLE()    SCI2.SCR.BIT.RIE=1

// when writing data be sure and clear the data empty
#define TX_DATA(data)   do {SCI2.TDR = data; SCI2.SSR.BIT.TDRE = 0;} while (0)

// Serial Buffer Error
BOOL rs232_error;

static UINT32   queue_size;

static UCHAR    *tx_queue;          // where the queue is
static UCHAR    *tx_head, *tx_tail; // track data in and out of the queue
static UCHAR    *tx_end;            // easy way to check for end of the queue

static UCHAR    *rx_queue;
static UCHAR    *rx_head, *rx_tail;
static UCHAR    *rx_end;

static UINT     rx_state;
static UINT     tx_state;
static UCHAR    TxNowData;

static UINT32   SerialOwner;



//inline
static UINT32
RxBytesFree(void)
{
    if (rx_tail > rx_head)
    {
        return (rx_tail - rx_head);
    }
    else
    {
        return (queue_size - (rx_head - rx_tail));
    }
}




UINT8
SerialError(void)
{
    // capture the interrupt error collection
    UINT8 error = rs232_error;

    // clear the background errors
    rs232_error = 0;

    if (SCI2.SSR.BIT.FER)
    {
        SCI2.SSR.BIT.FER = 0;   // Framing error
        error |= RS232_FRAMING_ERROR;
    }

    if (SCI2.SSR.BIT.ORER)
    {
        SCI2.SSR.BIT.ORER = 0;  // Overrun error

        error |= RS232_OVERRUN_ERROR;
    }

    if (SCI2.SSR.BIT.PER)
    {
        SCI2.SSR.BIT.PER = 0;   // Parity error

        error |= RS232_PARITY_ERROR;
    }

    return error;
}



void
SerialBegin(void)
{
    // setup the queue pointers
    tx_head = tx_tail = tx_queue;
    rx_head = rx_tail = rx_queue;

    rx_state = RX_XON;
    tx_state = TX_XON;

    // enable transmit and receive
    SCI2.SCR.BIT.TE = 1;
    SCI2.SCR.BIT.RE = 1;

    // clear the errors - can only clear them after TE and RE enabled
    (void)SerialError();

    // set the receive to not full and enable the receive interrupt
    SCI2.SSR.BIT.RDRF = 0;
    RXI_ENABLE();
}



void
SerialBreak(void)
{
    // nothing now
}



// stop transmit and receive
void
SerialEnd(void)
{
    // disable transmit and receive
    SCI2.SCR.BIT.TE = 0;
    SCI2.SCR.BIT.RE = 0;

    // TODO: Send break?
}





// INTERRUPT CONTEXT - Errors
// read and clear the errors
static void
RS232_ERROR_ISR(void)
{
    rs232_error = SerialError();
}



// INTERRUPT CONTEXT - Transmit Buffer has become empty
// No HandShake support
static void
RS232_TX_ISR_NHS(void)
{
    // any data to send? Had we received an XOFF or XON?
    if (tx_tail != tx_head)
    {
        // send the data
        TX_DATA(*tx_tail++);

        // wrap the pointer if is at the end of the queue
        if (tx_tail == tx_end)
        {
            // wrap pointer back to the start of the queue
            tx_tail = tx_queue;
        }
    }
    else
    {
        // no data - cancel the interrupt
        TXI_DISABLE();
    }
}


#if 0
// Save this for when the linker can do relaxing properly




// INTERRUPT CONTEXT - Transmit Buffer has become empty
static void
RS232_TX_ISR_HS(void)
{
    // sometimes need to send a special character like XON before other chars
    // in the queue - note: cannot send a zero now: 0x00
    if (TxNowData)
    {
        TX_DATA(TxNowData);
        TxNowData = 0;
    }
    // any data to send? Had we received an XOFF or XON?
    else if ((tx_state == TX_XON) && (tx_tail != tx_head))
    {
        // send the data
        TX_DATA(*tx_tail++);

        // wrap the pointer if is at the end of the queue
        if (tx_tail == tx_end)
        {
            // wrap pointer back to the start of the queue
            tx_tail = tx_queue;
        }
    }
    else
    {
        // no data - cancel the interrupt
        TXI_DISABLE();
    }
}



// Interrupt - receive buffer full with Software handshake
// receiving a ^S means stop sending - ^Q means start/continue
// Service it now - do the handshake
static void
RS232_RX_HS_ISR(void)
{
    UINT8 rs232data = SCI2.RDR;

    // the rx register is no longer full
    SCI2.SSR.BIT.RDRF = 0;
    // were we sent an XOFF? Means we should not send data out the TX anymore until we get a XON
    if (rs232data == ASCII_CTRL_S)  // ASCII_XOFF
    {
        tx_state = TX_XOFF;
    }
    else if (rs232data == ASCII_CTRL_Q)
    {
        tx_state = TX_XON;

        // no need to disable interrupts here - we are in an ISR: all irqs are disabled
        if (tx_tail != tx_head)
        {
            if (SCI2.SSR.BIT.TDRE)
            {
                TX_DATA(*tx_tail++);

                // wrap the pointer if is at the end of the queue
                if (tx_tail == tx_end)
                {
                    // wrap pointer back to the start of the queue
                    tx_tail = tx_queue;
                }
            }
        }
    }
    else if (RxBytesFree())
    {
        // add the data into the queue
        *rx_head++ = rs232data;

        if (rx_head >= rx_end)
        {
            rx_head = rx_queue;
        }
        // if not enough bytes free, send XOFF to avoid overflow
        if (RxBytesFree() < BUFFER_MARGIN && rx_state == RX_XON)
        {
            // send XOFF and record we are full
            if (SCI2.SSR.BIT.TDRE)
            {
                TX_DATA(ASCII_CTRL_S);
            }
            else
            {
                TxNowData = ASCII_CTRL_S;
            }

            // We are now RX_XOFF - more characters may appear because the
            // sender may keep sending while it is processing the XOFF we just
            // sent.
            rx_state = RX_XOFF;
        }
    }
}

#endif


// Interrupt - Receive-Buffer-Full without Software handshake or overflow check
static void
RS232_RX_NHS_ISR(void)
{
    // add the data into the queue
    *rx_head++ = SCI2.RDR;

    // the rx register is no longer full
    SCI2.SSR.BIT.RDRF = 0;

    // at the end of memory?
    if (rx_head == rx_end)
    {
        // wrap the pointer back to the beginning of the queue
        rx_head = rx_queue;
    }
}



// Read a byte from the queue with SW handshaking
StatusT
SerialRead(unsigned char *byte)
{
    // is there data in the queue?
    if (rx_tail == rx_head) return ST_EMPTY;

    // get the data from the queue
    *byte = *rx_tail++;

    // is the tail past the end of the queue - wrap it
    if (rx_tail >= rx_end)
    {
        rx_tail = rx_queue;
    }

    // Did we send an XOFF? - need to send an XON when ready to accept more
    // If handshaking is off then rx_state should always be RX_XON
    if ( (rx_state == RX_XOFF) && (RxBytesFree() > queue_size / 2) )
    {
        TXI_DISABLE();

        // if the data register is empty - use it now
        if (SCI2.SSR.BIT.TDRE)
        {
            TX_DATA(ASCII_CTRL_Q);
        }
        else
        // the transmit buffer is full and is waiting to be transfered
        // to the transmit shift register
        {
            TxNowData = ASCII_CTRL_Q;
        }

        TXI_ENABLE();

        // the device sending to us is free to send
        rx_state = RX_XON;
    }

    return ST_OK;
}



StatusT
SerialSend(UCHAR data)
{
    StatusT status = ST_OK;

    // Disable the transmit empty buffer interrupt - TXI
    // The ISR may be transfering data during this function and the empty bit
    // might assert during this function. If a byte is not put in here, then
    // the ISR will execute and service an empty queue. So turn off the
    // interrupts so that when it is turned on it will find a byte in the queue.
    TXI_DISABLE();

    // bypass the buffer and put the data directly into the TDR if empty
    if (SCI2.SSR.BIT.TDRE)
    {
        TX_DATA(data);
    }
    else    // add to the buffer - if there is room
    {
        UCHAR *tx_next = tx_head + 1;

        // wrap the pointer for the next entry if at the end
        if (tx_next >= tx_end)
        {
            // wrap pointer back to the start of the queue
            tx_next = tx_queue;
        }

        // check for overflow - is the buffer full?
        if (tx_next == tx_tail)
        {
            // no room
            status = ST_FULL;
        }
        else    // there is room for the char
        {
            // add data to the queue
            *tx_next = data;
            tx_head = tx_next;
        }
    }

    TXI_ENABLE();

    return status;
}



// Note: be sure all data has been processed before closing
StatusT
SerialClose(void)
{
    // turn off the interrupts
    SCI2.SCR.BIT.TE = 0;
    SCI2.SCR.BIT.RE = 0;

    // TODO: confirm the right owner
    SerialOwner = 0;

    // stop serial module
    MSTPCR.BIT._SCI2 = 1;

    // free the queues
    if (tx_queue)
    {
        free(tx_queue);
        tx_queue = 0;
    }

    if (rx_queue)
    {
        free(rx_queue);
        rx_queue = 0;
    }

    return ST_OK;
}



void static
SerialBaud(BaudT baud)
{
    // set the baud rate
    switch (baud)
    {
        case RS232_BAUD_2400:
            SCI2.BRR = BAUD_REG(2400);
            break;

        case RS232_BAUD_4800:
            SCI2.BRR = BAUD_REG(4800);
            break;

        case RS232_BAUD_9600:
            SCI2.BRR = BAUD_REG(9600);
            break;

        case RS232_BAUD_19200:
            SCI2.BRR = BAUD_REG(19200);
            break;

        case RS232_BAUD_38400:
            SCI2.BRR = BAUD_REG(38400);
            break;

        default:    // cybiko sdk console default
        case RS232_BAUD_57600:
            SCI2.BRR = BAUD_REG(57600);
            break;

        case RS232_BAUD_115200:
            SCI2.BRR = BAUD_REG(115200);
            break;

    }
}



static void
SerialFormat(FormatT format)
{
    // set the format
    switch (format)
    {
        default:
        case RS232_FORMAT_8N1:
            SCI2.SMR.BIT.CHR = 0;   // 8 bits
            SCI2.SMR.BIT.STOP = 0;  // one stop bit
            SCI2.SMR.BIT._PE = 0;   // parity disabled
            break;

        case RS232_FORMAT_8D1:
            SCI2.SMR.BIT.CHR = 0;   // 8 bits
            SCI2.SMR.BIT.STOP = 0;  // one stop bit
            SCI2.SMR.BIT._PE = 1;
            SCI2.SMR.BIT.OE = 1;    // 0 is even, 1 is odd
            break;

        case RS232_FORMAT_8E1:
            SCI2.SMR.BIT.CHR = 0;   // 8 bits
            SCI2.SMR.BIT.STOP = 0;  // one stop bit
            SCI2.SMR.BIT._PE = 1;
            SCI2.SMR.BIT.OE = 0;
            break;

    }

}



StatusT
SerialOpen(BaudT baud, FormatT format, UINT32 size)
{
    queue_size = size;
    if (queue_size < QUEUE_SIZE_MIN)
    {
        queue_size = QUEUE_SIZE_MIN;
    }


    tx_queue = (UCHAR *)malloc(queue_size);
    if (tx_queue == 0)
    {
        return ST_MEMORY;
    }


    rx_queue = (UCHAR *)malloc(queue_size);
    if (rx_queue == 0)
    {
        free(tx_queue);
        return ST_MEMORY;
    }

    // set the queue endings
    tx_end = tx_queue + queue_size;
    rx_end = rx_queue + queue_size;

    // TODO: Assign the owner
    SerialOwner = (UINT32)tx_queue;

    // turn off the stop for serial module - cant use registers until then
    MSTPCR.BIT._SCI2 = 0;

    // disable transmit and receive before
    // applying the settings
    SCI2.SCR.BIT.TE = 0;
    SCI2.SCR.BIT.RE = 0;

    // always use this fast clock selection - 0
    SCI2.SMR.BIT.CKS = 0;

    SerialFormat(format);

    // make sure other things are correct
    SCI2.SMR.BIT.CA = 0;    // set asyncronous mode
    SCI2.SMR.BIT.MP = 0;    // set multiprocessor mode disabled
    SCI2.SCR.BIT.CKE = 0;   // set config clock pin as IO - use internal clock

    // Compute and set the baud rate
    SerialBaud(baud);

    // Need to delay for one bit length - about 50uS for 19200


    // set the handshake mode
    if (1 /*hs == HANDSHAKE_OFF*/)
    {
        TXI2_VECTOR = (UINT32)RS232_TX_ISR_NHS;
        RXI2_VECTOR = (UINT32)RS232_RX_NHS_ISR;     // No HandShaking
    }
    else    // handshake ON
    {
        //TXI2_VECTOR = (UINT32)RS232_TX_ISR_HS;
        //RXI2_VECTOR = (UINT32)RS232_RX_HS_ISR;     // HandShaking
    }

    // cant reset error flags until RE is enabled
    ERI2_VECTOR = (UINT32)RS232_ERROR_ISR;

    return ST_OK;
}



