/*
**************  F O N T S - open and close

Fonts have to be stored in a consistant way and converted to a bitmap/s

There are 2 ways to render fonts: vector and bitmap
There are 4 types of fonts: scalable, fixed, proportional and fast.

Design Option 1:
Glyphs are stored in an efficient way and converted to an efficient storage
during open. Stored bits example = 4 * 8 * 100 = 3200 / 8 = 400
Converted bits = 8 * 8 * 100 = 6400 / 8 = 800
Pros: Fast rendering
Cons: More dynamic storage needed. Converting takes time - caching needed.

Design Option 2:
Compact Glyphs are the font. Dont do any conversion.
Stored bits example = 4 * 8 * 100 = 3200 / 8 = 400 + overhead
Pros: Least memory needed. No conversion or caching needed
Cons: Slower rendering.

Design Option 3:
Font is stored in render preferred format.
Stored bits example = 8 * 8 * 100 = 6400 / 8 = 800 + overhead
Pros: No conversion needed. Fast Rendering.
Cons: More static storage needed

*/


#include <lowlevel.h>
#include <gfx.h>



// Max fonts that can be open at one time
#define MAX_FONTS   10



typedef struct
{
    FontIdT     id;
    FontT       font;
    UINT16      used;
} FontCacheEntryT;


// The font structure to complete the data type in the header
struct font
{
    PixelmapT   *glyph;     // pointer to list of maps of glyphs
    FontAttrT   attributes; // proportional or fixed or scaled or fast
    UCHAR       first;      // what ASCII value is the first char
    UCHAR       last;       // last ASCII value
    UCHAR       maxw;       // width of widest glyph
    UCHAR       maxh;       // maximum heigth - decenders?
};


typedef struct
{
    FontIdT id;
    GlyphsT *glyphs;
} name;



static FontCacheEntryT FontCache[MAX_FONTS];



static FontT
FontCached(const FontIdT id)
{
    int i;

    // look for the open font
    for (i = 0; i < MAX_FONTS; i++)
    {
        if (FontCache[i].id == id)
        {
            // font is already open
            // record another usage
            FontCache[i].used++;

            // return the pointer to the requested font
            return FontCache[i].font;
        }
    }

    return (FontT)NULL;
}



// turn the stored font found at i into a useable font at j
// A stored font is stored at glyphs. A useable font is stored
//   as pixmaps ready to be put onto a surface
//

static FontT
FontConvert(const GlyphsT * StoredFont)
{
    FontT   NewFont;
    void    *FontMemory;
    UINT    glyphcount;
    UINT    glyphbytes;
    UINT    pixelmapbytes;

    // number of glyphs to convert
    glyphcount = StoredFont->last - StoredFont->first + 1;

    pixelmapbytes = glyphcount * sizeof(PixelmapT);

    //max amount of bytes needed = w * h * count
    glyphbytes = ((StoredFont->maxw + 3) / 4) * StoredFont->maxh * glyphcount;

    FontMemory = malloc(pixelmapbytes + glyphbytes);

    // did we get memory?
    if (FontMemory == NULL) return NULL;

    // split up the font memory - remember to use FREE just for the one
    // memory block at font->glyph later
    NewFont->glyph  = (PixelmapT *)FontMemory;

    // skip the
    UINT8 *PixelmapData = (UINT8 *)FontMemory + pixelmapbytes;
    NewFont->type   = StoredFont->type;
    NewFont->first  = StoredFont->first;
    NewFont->last   = StoredFont->last;

    // The stored format has a loose format
    // There is a fixed header, but the data is inline and variable len
    UBYTE *StoredGlyph = StoredFont->glyphs;

    // The height is always the same for the entire font - TODO really?
    NewFont->maxh = StoredFont->maxh;


    // Now each glyph has to be constructed and filled in
    for (glyph = NewFont->first; glyph <= NewFont->last; glyph++)
    {
        // Convert a stored glyph into a NewFont glyph: pixmap

        // First do the header
        GlyphPixmap->width = StoredGlyph->woff + StoredGlyph->w;
        GlyphPixmap->height = StoredGlyph->hoff + StoredGlyph->h;

        // convert a pixel width into how many bytes are needed
        // Bits stored need to translate into 2bits per pixel
        // So, 6 pixels wide = 12 bits -> 2 bytes
        GlyphPixmap->stride = (GlyphPixmap->width + 3) >> 2;

        // assign the pointer for pixmap
        GlyphPixmap->image = PixelmapData;

        // for all of the data in the stored glyph
        // skip the header and start converting the stored data.
        // 'stride' bytes per row are needed
        // number of bytes used = stride * height
        // Do rows
        //  PixmapData[xxx]

        // move to open space for next glyph data to go in pixelmap
        PixelmapData += (GlyphPixmap->stride * GlyphPixmap->height);

        // next glyph pixelmap to fill it
        GlyphPixmap++;

    }


}




static GlyphsT *
GlyphsStored(const FontIdT id)
{
    for (i=0; i<KNOWN_FONTS; i++)
    {
        // did we find the requested font?
        if (FontsKnown[i].id == id)
        {
            return FontsKnown[i].glyphs;
        }
    }

    return NULL;
}



static void
FontCache(const FontT font, const FontIdT id)
{
    int j;

    // Is there room to add a font?
    // find a place for the font
    for (j = 0; j < MAX_FONTS; j++)
    {
        if (FontCache[j].used == 0)
        {
            // unused section found
            FontCache[j].font = font;
            FontCache[j].id = id;
            FontCache[j].used++;
        }
    }
}



FontT
GfxFontOpen(FontIdT id)
{
    FontT   NewFont;
    GlyphsT *StoredGlyphs;

    // check if the font is cached
    NewFont = FontCached(id);
    if (NewFont) return NewFont;

    // does this id match any stored font glyphs?
    StoredGlyphs = GlyphsStored(id);
    if (NULL == StoredGlyphs) return NULL;

    // convert the stored glyphs into a font
    NewFont = FontConvert(StoredGlyphs);

    // did it convert OK?
    if (NULL = NewFont) return NULL;

    // must work
    FontCache(NewFont);

    return NewFont;
}



void
GfxFontClose(FontT font)
{
    int j;

    // check all entries for the font
    for (j = 0; j < MAX_FONTS; j++)
    {
        if (FontCache[j].font == font)
        {
            // is everyone done using this font?
            if (--FontCache[j].used == 0)
            {
                // release the memory for this font
                free(font->glyph);

                // remove the font from the cache
                FontCache[j].font = NULL;
            }

            break;
        }
    }
}



