// CYBIKO REBORN OS - CyBorn

#ifndef _CYBORN_H_
#define _CYBORN_H_


#define ISR_ENTER {SwitchTask = 0;}
#define ISR_EXIT  {if (SwitchTask) ContextSwitch(SwitchTask);}


#define IS_HIGHER_THAN_CURRENT(task) \
    ((task)->priority < CurrentContext->priority)

#define IS_LOWER_THAN_CURRENT(task) \
    ((task)->priority > CurrentContext->priority)



#define LIST_UNLINK(task)               \
        task->prev->next = task->next;  \
        task->next->prev = task->prev;


#define LIST_LINK_FRONT(task, head)     \
            task->next = head->next;    \
            task->prev = head->prev;    \
            head = task;                \
            task->prev->next = task;


#define LIST_LINK_BACK(task, head)          \
            task->next = head->prev->next;  \
            task->prev = head->prev;        \
            head->prev->next = task;        \
            head->prev = task;

#define IS_HIGH_PRIORITY(task) (task->priority == TASK_PRIORITY_HIGH)


#define INTERRUPTS_ENABLE   asm("andc #0x3f,ccr");      // Enable interrupts
#define INTERRUPTS_DISABLE  asm("orc  #0xc0,ccr");      // Mask interrupts


#define CONTEXT_SWITCH      asm("bnot #0,@0x6f:8");     // Trigger IRQ6




// things for the header file
typedef volatile unsigned long CybornTickT;




//#define TASK_PRIORITY_CRIT  0x00
//#define TASK_PRIORITY_HIGH  0x01
//#define TASK_PRIORITY_LOW   0x02
//#define TASK_PRIORITY_IDLE  0x03



//////////////////////////////////////////////////////////////////////////
//////////////  T A S K S
//////////////////////////////////////////////////////////////////////////
typedef enum
{
    TASK_PRIORITY_SYSTEM,
    TASK_PRIORITY_TOP,
    TASK_PRIORITY_HIGH,
    TASK_PRIORITY_MED,
    TASK_PRIORITY_LOW,
    TASK_PRIORITY_IDLE,
    TASK_PRIORITIES,        // last entry for how many are in the list
} PriorityT;


#define TASK_NONE   0


typedef unsigned char   *AddressT;

typedef void (*EntryT)(void *parameter);



typedef struct task
{
    // need this for linking tasks in and out of lists
    struct task     *next;          // need to stay in front
    struct task     *prev;

    // space to save the stackpointer when switching contexts
    AddressT        stackpointer;   // holds pointer into stackframe

    // used when put on the timer list
    CybornTickT     ticks;

    // what function to call and parameter to use
    EntryT          entry;          // entry point for task
    void            *parameter;     // parameter for task

    // priority of this task
    PriorityT       priority;       // priority helps sort tasks

    void            *schedule;      // let the scheduler attach something
    short           stackframe[];   // stack space is extra in task
} TaskT;



// Task API

TaskT *TaskCreate(  EntryT          entry,
                    unsigned long   StackSize,
                    void            *parameter,
                    PriorityT       priority);


//////////////////////////////////////////////////////////////////////////


#if 0
typedef struct queueblockedt
{
    struct queueblockedt    *next;
    struct queueblockedt    *prev;

    CybornQueueT            *queue;     // queue

    // message pointer
    // size
} QueueBlockedT;
#endif




// Generic data type for a list
// Use:
//  CybornSemT sem;
//  sem->waiting = (TaskT *)&sem->list;
// Now the list is not empty and always points to something
// This makes linking and unlinking faster because we don't have to test for
// head pointing, etc
typedef struct
{
    struct task *next;
    struct task *prev;
} TaskListT;





// a mutex allows only one process to enter
typedef struct
{
    TaskT                   *next;      // this is the list
    TaskT                   *prev;
    volatile unsigned char  *flag;      // points to onchip ram used for TAS
    TaskT                   *owner;     // task that owns this
    TaskListT               list;
} CybornMutexT;



typedef struct
{
    TaskT                   *next;      // this is the list
    TaskT                   *prev;
    volatile unsigned       flag;       // memory protected by irq
} CybornSemT;




typedef enum
{
    CYBORN_STATUS_OK,           // all is well
    CYBORN_STATUS_TIMEOUT,
    CYBORN_STATUS_NULL,         // NULL pointer not expected
    CYBORN_STATUS_FULL,
    CYBORN_STATUS_NOINIT,       // the resource is not initialized
    CYBORN_STATUS_EMPTY,        // empty
    CYBORN_STATUS_BLOCKED,      // blocked
    CYBORN_STATUS_MEMORY,       // not enough memory
    CYBORN_STATUS_RESOURCE,     // no such resource
    CYBORN_STATUS_OWNER,        // not the owner
    CYBORN_STATUS_NOTFOUND,     // couldn't find it
} CybornStatusT;



//////////////////////////////////////////////////////////////////////////
//////////////  Q U E U E S
//////////////////////////////////////////////////////////////////////////
typedef char QueueT;


typedef struct
{
    QueueT          *head;
    QueueT          *next;
    QueueT          *end;
    CybornSemT      sem;
    QueueT          *queue;
} CybornQueueT;



////   Character Queues

CybornStatusT
CyqInit(CybornQueueT *queue, unsigned long size);


void
CyqDestroy(CybornQueueT *queue);


CybornStatusT
CyqWait(    CybornQueueT    *queue,     // which queue to wait on
            char const      *message,   // where to put message
            unsigned long   *count);    // size of message


CybornStatusT
CyqAdd( CybornQueueT  *queue,     // queue
        char          character); // add ONE character at a time




//////////////////////////////////////////////////////////////////////////
//////////////  M U T E X E S - Critical regions
//////////////////////////////////////////////////////////////////////////

// Using Critical regions
// Protects data from corruption
//  status = CriticalInit(&database);  // check the status
//  CriticalEnter(&database);
//  do database stuff
//  CriticalExit(&database);
//  // if all done with the sem then toss it:
//  CriticalDestroy(&database);
//



CybornStatusT
CriticalTry(CybornMutexT *mutex);


void
CriticalEnter(CybornMutexT *mutex);


void
CriticalExit(CybornMutexT *mutex);


CybornStatusT
CriticalInit(CybornMutexT *mutex);


void
CriticalDestroy(CybornMutexT *mutex);



//////////////////////////////////////////////////////////////////////////
//////////////  S E M A P H O R E S - Signal
//////////////////////////////////////////////////////////////////////////


void
SignalBGive(CybornSemT *semaphore);


void
SignalCGive(CybornSemT *semaphore);


void
SignalBIsrGive(CybornSemT *semaphore);


void
SignalCIsrGive(CybornSemT *semaphore);


void
SignalWait(CybornSemT *semaphore);


CybornStatusT
SignalTry(CybornSemT *semaphore);


//CybornStatusT
//SignalInit(CybornSemT *semaphore, int initial, SemTypeT type);


void
SignalDestroy(CybornSemT *semaphore);

//////////////////////////////////////////////////////////////////////////



//////////////////////////////////////////////////////////////////////////
//////////////  S C H E D U L E R
//////////////////////////////////////////////////////////////////////////


// wake the task - add it to ready queue
void inline ScheduleWake(TaskT *task);

// Delete the task from wherever it is
void inline ScheduleDelete(TaskT *task);

// The task will be moved from running to a waiting queue
void inline ScheduleWait(TaskT *index);

// remove the current task from Q and figure out the next one to run
void inline ScheduleRemove(void);

void inline ScheduleDisable(void);

void inline ScheduleEnable(void);

// register the task with the scheduler
void ScheduleRegister(TaskT *task);

//////////////////////////////////////////////////////////////////////////



int     TimerReady(void);





#endif  // stop extra includes
