#include "dataflash.h"
#include "cyborn.h"

#include "dtc.h"



void DataFlashSend(unsigned char byte)
{
    // wait until the data buffer is empty
    while (SCI1.SSR.BIT.TDRE == 0) ;

    // send the byte by writing to the transmit data register
    SCI1.TDR = byte;
}




/*
The following procedure should be used for serial data reception:
When changing the operating mode from asynchronous to clocked synchronous, be
sure to check that the ORER, PER, and FER flags are all cleared to 0.
The RDRF flag will not be set if the FER or PER flag is set to 1, and neither
transmit nor receive operations will be possible.
*/

static int
DataFlashBusy(void (*function)(void))
{
    char status;

    // turn off the receive
    SCI1.SCR.BIT.RE = 0;

    // Write the status opcode out the SCI
    // ChipSelect = 0
    SCI1.SSR.BIT.ORER = 0;     // clear the error flag
    SCI1.SCR.BIT.TE = 1;       // enable transmit

    // status please...
    DataFlashSend(DATAFLASH_OP_STATUS);

    // wait until the data has been transmitted
    while (SCI1.SSR.BIT.TEND == 0) ;


    // Read the status byte from the data flash

    // change the port for the clock to an output port for manual operation
    // because we can read the status bit directly

    // change the clock to internal/off so we can read the status bit at the
    // port

    // monitor the input pin until the wait is over
    while (!P3.BIT.B3)
    {
        if (function)
        {
            function();
        }
    }

}



// interrupt for SCI 1
static void Sci1Interrupt(void)
{
    SignalGive(DtcAccessFinished);
}



#define SCI1_DTC_VECTOR 0xFFFBD0


// read 264 bytes at a time (256 + 8)
// can read as fast as we want
void DataFlashReadPage(unsigned page, char *buffer)
{
    DtcT *dtc;

    dtc = (DtcT *)SCI1_DTC_VECTOR;

    // only allow one process in here at a time
    CriticalEnter(DataFlashAccess);

    // set the data direction to MSB first
    SCI1.SCMR.BIT.SDIR = SCI_CLOCK_MSB;

    // setup the DTC for the read
    dtc->LONG1.Address = &SCI1.RDR;
    dtc->LONG1.Bytes.mode = 0x20;
    dtc->Cra = DATAFLASH_PAGE;

    dtc->LONG2.Address = buffer;
    dtc->LONG2.Bytes.mode = 0x00;
    //dtc->Crb = not used;


    // wait until flash is ready - call function: ScheduleYield while waiting
#ifdef _WITH_CYBORN_
    DataFlashBusy(ScheduleYield);
#else
    DataFlashBusy(0);
#endif

    // Start the transfer - disable preemption
    // Since we are going to be blocking for a long time soon...
#ifdef _WITH_CYBORN_
    SchedulePreemption(SCHEDULE_PREEMPT_SUSPEND);
#else

    // setup the SCI 1 to talk to the flash
    // send the bytes to the flash
    DataFlashSend(DATAFLASH_OP_READMAIN);
    DataFlashSend(page >> 7);
    DataFlashSend(page << 1);
    DataFlashSend(0x00);

    // extra 4 bytes
    DataFlashSend(0x00);
    DataFlashSend(0x00);
    DataFlashSend(0x00);
    DataFlashSend(0x00);

    // wait here til transmission is complete
    while (SCI1.SSR.BIT.TEND == 0) ;


    // read the bytes out
    // setup the SCI1 for rx
    // The DTC will now transfer all the bytes we programmed into the SCI1 DTC


    // enable preemption - transfer runs in the back ground
#ifdef _WITH_CYBORN_
    SchedulePreemption(SCHEDULE_PREEMPT_RESUME);
#endif

    // suspend here until the reading is finished
    // The signal will be given by a DTC interrupt triggered by end of transfers
#ifdef _WITH_CYBORN_
    SignalWait(DtcAccessFinished);
#else
    while (DTCnotfinished) ;
#endif

    // done with this access
    CriticalExit(DataFlashAccess);
}



// only write a page at a time
// all pages are 264 bytes
void DataFlashWritePage(unsigned page, char *buffer)
{
    DtcConfigT DtcConfig;
    char OpcodeBuffer[8];


    // One process at a time
    CriticalEnter(DataFlashAccess);

    // setup the DTC for the opcode and address send
    // this DTC will chain to the next
    DtcConfig.SrcCount = 8;
    DtcConfig.SrcAddress = OpcodeBuffer;

    DtcConfig.DstCount = 1;
    DtcConfig.DstAddress = &SCI1->TDR;

    DtcConfig.Mode = DTC_MODE_CHAIN | ;
    __DtcSetup(0xFFFBD0, &DtcConfig);



    // setup the next transfer
    DtcConfig.SrcCount = 1;
    DtcConfig.SrcAddress = &SCI1->RDR;

    DtcConfig.DstCount = DATAFLASH_PAGE;
    DtcConfig.DstAddress = buffer;

    DtcConfig.Mode = 0;
    __DtcSetup(0xFFFBD0 + 0x0C, &DtcConfig);


    // setup the SCI 1 - already set by init?
    __SciSetup(setup);


    // cant operate on flash while its busy
    // This function will call the given function until ready
    DataFlashBusy(ScheduleYield);

    // Start the transfer
    // Enable the correct chip select...

    // suspend here until the DTC is finished
    SignalWait(DtcAccessFinished);

    // Disable the chip select


    // done with the write request - the main memory is busy
    CriticalExit(DataFlashAccess);
}




// setup
void DataFlashInit(void)
{
    // connect the vector for DTC
    // set the SCI1 BBR1
    // set clock divisor
    // connect the ISR for SCI1
    // initialize the Signal
    // initialize the Mutex for the Critical region
}

