#include "cyborn.h"

// There will be two priorities: high and low represented by a double linked
//// list. The front of the list will have the highest priority tasks.
//// Round robin scheduling will be implemented by following the link of the
//// current running task to the next. If the next is the same priority, it will
//// run next, else the next task will be the first on the list.





/*
Context Frame


When the interrupt occurs, the CCR and PC are pushed on the stack

/////////////////////////////////////////////////////////////////////
Example:
stackpointer before = xx1000
stackpointer after  = xx0FFC
 ======
|  CCR |  xx0FFC   <---- SP after interrupt
 ======
|  PC  |  xx0FFD
|      |  xx0FFE
|      |  xx0FFF
 ======
/////////////////////////////////////////////////////////////////////


\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
Hard coded vector to hard coded monitor location:
IRQ0 vector 16 - 00 00 11 2c
      40:   00 00           00 00
      42:   11 2c           11 2c


The cybiko has a hard interrupt service routine with a soft vector. The hard
service is in ROM and cannot be changed. It points to a vector in RAM that can
be changed

Here is where the vector points: inside the ROM
    112c:	01 30 6d f0 	01 30 6d f0       stm.l er0-er3,@-sp
    1130:	01 20 6d f4 	01 20 6d f4       stm.l er4-er6,@-sp
    1134:	01 00 6b 02 	01 00 6b 02 ec 14 mov.l @0xec14:16,er2
    1138:	ec 14
    113a:	5d 20       	5d 20             jsr   @er2
    113c:	01 20 6d 76 	01 20 6d 76       ldm.l @sp+,er4-er6
    1140:	01 30 6d 73 	01 30 6d 73       ldm.l @sp+,er0-er3
    1144:	56 70       	56 70             rte

Change the vector at 0xFFEC14 to point to a desired ISR


\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
IRQ0 and the timer int look similar
Before the return from the routine, the stack is like this:

EXTRA    <----- SP after the JSR to RAM vector
PC2     - address of next instruction (00113c in the case of IRQ0)
PC1
PC0
er6-3    <----- SP after interrupt
er6-2
er6-1
er6-0
er5-3
er5-2
er5-1
er5-0
er4-3
er4-2
er4-1
er4-0
er3-3
er3-2
er2-1
er2-0
er1-3
er1-2
er1-1
er1-0
er0-3
er0-2
er0-1
er0-0
CCR
PC2     - where to start/resume execution
PC1
PC0
         <----- SP before interrupt IRQ6 triggered by bnot #0


When the scheduler is running - this is what the stack will need to look like
so that the task switch will point to the stack at the same place for each
task that needs to be continued. The Scheduler will load the stack pointer for
the selected task and then do an rts which will pop the PC off the stack,
return to the monitor rom section, load the registers off the stack, do an
rte which will complete the context switch by restoring the flags and
program counter. Here is the problem to solve: what to do about the FIRST time
that this task should run? Answer: since the instruction is an rts, the only
thing that needs to be on the stack is the vector for the entry point and move
the stackpointer to the entrypoint on the stack.


EXTRA
PC2   - entry point for task
PC1
PC0

This is done in an interrupt context by the scheduler. The scheduler doesn't
care about the registers, but the interrupt context needs to be adjusted
to be enabled at the beginning of each task. After the task has been properly
and interrupts enabled, the context switch where it does ALL of the registers
including CCR will be the normal activity.


*/



// point this to the current queue of ready tasks - may have to lock this if
// getting it outside of the system tick - like during blocking
static TaskT *CurrentContext;
static TaskT *NextContext;


TaskListT TaskReadyList;
TaskListT *pTaskReadyList = &TaskReadyList;



#define TIMING_SOURCE_WATCHDOG
#define CYBIKO_SYSTEM_CLOCK         11059200

// for changing the watchdog register
// The watchdog counts from 0 to 255 and then causes interrupt
// Interrupt freq = sysclock / DIV / 256
//                                          //    Hz       period
#define WATCHDOG_DIV_2              0x00    // 21600        46uS
#define WATCHDOG_DIV_64             0x01    //   675        1.5mS
#define WATCHDOG_DIV_128            0x02    //   337.5      3.0mS
#define WATCHDOG_DIV_512            0x03    //    84.375
#define WATCHDOG_DIV_2048           0x04    //    21.09375
#define WATCHDOG_DIV_8192           0x05    //     5.2734375
#define WATCHDOG_DIV_32768          0x06    //     1.318359375
#define WATCHDOG_DIV_131072         0x07    //     0.32958984375

#define WATCHDOG_DIV_MASK           0x07

#define WATCHDOG_DIV                WATCHDOG_DIV_2048


//                                          clocks per timer tick
#define TIMER0_DIV_0              0x00    // stopped
#define TIMER0_DIV_8              0x01    // 1382400
#define TIMER0_DIV_64             0x02    //  172800
#define TIMER0_DIV_8192           0x03    //    1350

#define TIMER0_DIV_MASK           0x07

#define TIMER0_DIV                TIMER0_DIV_64



#define TICKS_PER_SECOND            (800)
#define SECONDS_TO_TICKS(s)         ((s) * TICKS_PER_SECOND)
#define MILLISECONDS_TO_TICKS(ms)   ((((ms) * TICKS_PER_SECOND) + 500) / 1000)




///////////////////////////////////////////////////////////// PRIORITIES

///////////////////////////////////////////////////////////// PRIORITIES
// These queues are for priorities and scheduling



typedef void (*IsrEventT)(void *parameter);






///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////   PERIODIC
/////////////////////////////////////////////////////////////   EVENT
#ifdef USE_PERIODIC_EVENT

// Periodic Event - checked at every event
typedef struct
{
    IsrEventT       event;
    void            *param;
    CybornTickT     period;
    CybornTickT     timer;
} PeriodicEventT;


// ONCHIP...
static PeriodicEventT *PeriodicEvent;


// Use like this:
// PeriodicEvent(dothis, MILLISECONDS_TO_TICKS(80));
// PeriodicEvent = event;
// -- to stop: PeriodicEvent = 0;
// API

// clear by setting event to null
// ie PeriodicEventSet(0, 0, 0);
void
PeriodicEventSet(IsrEventT event, void *parameter, CybornTickT period)
{
    static PeriodicEventT event;

    // doing it this way means we don't have to disable ints...
    PeriodicEvent = 0;

    // fill in the struct
    event.event     = event;
    event.param     = parameter;
    event.period    = period;
    event.timer     = 0;

    // point to ready-to-go struct
    PeriodicEvent = &event;
}

#endif
////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////
// Internal storage


void inline
ScheduleDisable(void)
{
    // Disable interrupts
}



void inline
ScheduleEnable(void)
{
    // Enable interrupts
}



CybornTickT CybornTicksPerSecond(void)
{
    return TICKS_PER_SECOND;
}



//////////////////////////////////////////////////////////////////////////////
// Call this with the task that needs to be added to the prioritized ready list
// from inside an ISR
TaskT *
ScheduleIsrReady(TaskT *task)
{
    // REMOVE task from other blocking list
    LIST_UNLINK(task);

    // ADD task to ready to run list
    if (task->priority == TASK_PRIORITY_HIGH)
    {
        // Add to the FRONT of the ready list
        LIST_LINK_FRONT(task, pTaskReadyList->next);

        // if the task has a higher priority it should be run
        if (IS_HIGHER_THAN_CURRENT(task))
        {
            return task;
        }
    }
    else  // low - lower priority
    {
        // Add to the BACK of the list
        LIST_LINK_BACK(task, pTaskReadyList->next);
    }

    // return null for no switch
    return (TaskT *)0;
}




//////////////////////////////////////////////////////////////////////////////
// Call this with the task that needs to be added to the prioritized ready list
// DO NOT ADD A TASK TWICE!
inline void
ScheduleWake(TaskT *task)
{
    INTERRUPTS_DISABLE;

    // remove the task from the list it is on
    LIST_UNLINK(task);

    // does this new task made ready have a higher priority than the current one
    if (task->priority == TASK_PRIORITY_HIGH)
    {
        // link to the front where the priority is highest
        LIST_LINK_FRONT(task, pTaskReadyList->next);
    }
    else
    {
        // link to the back where the priority is lowest
        LIST_LINK_BACK(task, pTaskReadyList);
    }

    INTERRUPTS_ENABLE;
}




#if 0
// demo for SemTake implementation
void
SemTake(SemT *sem)
{
    ...

    // puts the return address on the stack and makes another task current

    ScheduleBlock(sem->blocked);
}
#endif




// Moves current task from the running list it is on, to another list
// A task must be inserted in a list someplace!
// A different task will be run for sure
void
ScheduleBlock(TaskT *blocked)
{
    INTERRUPTS_DISABLE;

    // hold the next task, the task pointers are going to change
    NextContext = CurrentContext->next;

    // REMOVE CurrentContext from the ready list
    LIST_UNLINK(CurrentContext);

    // Now transfer the current task to the blocked list
    // The list could be for sems, timer, etc
    // Add to priority queue - determine priority
    if (IS_HIGH_PRIORITY(CurrentContext))
    {
        LIST_LINK_FRONT(NextContext, blocked);
    }
    else
    {
        LIST_LINK_BACK(NextContext, blocked);
    }


    INTERRUPTS_ENABLE;

    // returns later when context switches back
    return;
}



// Puts the task on the suspend list
// NOTE: Any task put here will not be run again until put on the ready Q
// A task could suspend itself... but cant restart itself
//   ScheduleSuspend(task);
//  later: by another task
//   ScheduleNow(task);

#if 0
void
ScheduleSuspend(TaskT *task)
{
    TaskT *index;

    index = (TaskT *)&TaskSuspendList);

    // remove task from the ready list and send to the suspend list
    ScheduleRemove(task, index);
}
#endif




// Yields to the next process to run in the ready priority queue
// Call this from userland only - the contextswitch only works there
// Can't have any local variables because the stack needs to be aligned
void
ScheduleYield(void)
{
    // mask all interrupts:
    INTERRUPTS_DISABLE;

    // is the next priority too low?
    if (CurrentContext->next->priority > pTaskReadyList->next->priority)
    {
        // switch to the head of the list
        NextContext = TaskReadyList.next;
    }
    else
    {
        // point to the next task
        NextContext = CurrentContext->next;
    }

    // only switch context if different task to run
    if (NextContext != CurrentContext)
    {
        // do it - only happens after the interrupts are enabled
        CONTEXT_SWITCH;
    }

    // enable interrupts - as soon as enabled - the IRQ6 should fire
    INTERRUPTS_ENABLE;

    return;
}





// what should the idle task do?
// system stats?
// Its the first one to start so do initialization
static void
IdleTask(void *nothing)
{
    // Do things here to initialize
    // Most important is that the stack for this task is filled with right vals
    // 1> Set Watchdog to clock divide and repeat output
    // 2> Set timer vector in onchip RAM to point to ScheduleSlice
    // 3> Enable the Watchdog timer interrupt - next interrupt will
    //    schedule the high priority task - if there is one

    for (;;)
    {
        // Do something idle here - never sleep or be taken off the ready Q
        // Check for screensaver?
        // Should we power down?
        // Preemption will break in here and choose a task with higher priority

    }
}


#if 0
// Example starting
TaskT *begin;

// set up the list for tasks
begin = SchedulerInit();
if (begin)
{
    // start all of the tasks that we want
    // NewTask.... ScheduleAdd.... repeat

    SchedulerStart(begin);

#endif


// all tasks must servive a task switch: all regsiters, pc to return, jsr to
//// new vector
#define TASK_STACK_SIZE_MIN  72

#define IDLE_TASK_STACK_SIZE (TASK_STACK_SIZE_MIN + 20)


// ISRs are off
// Initializes the scheduler with an idle task - must have at least one task
// Returns pointer to idle task - or 0 if error - the scheduler is ready to run
TaskT *SchedulerInit(void)
{
    static TaskT    *idle;  // static cause it must persist
    extern void     ContextSwitch(void);

    // mask IRQ6 - set IRQ6 to trigger on flip
    // Set port G.0 DDR to output so that we can trigger it internally since
    //// we dont have a trap vector available

    asm("orc #0xc0,ccr");       // Mask interrupts
    asm("bset #4, @0x2c:8");      // ISCRH - Trigger on both edges.
    asm("bset #5, @0x2c:8");
    
    // ISR - Clear interrupt status of irq6
    asm("bclr #6, @0x2f:8");

    // IER - Enable IRQ6
    asm("bset #6, @0x2e:8");


    asm("mov.b #0x1f,r0l");      // Enable PGDDR pin 0
    asm("mov.b r0l,@0xfebf:16");
    // IRQ6 can now be software generated via - BNOT #0,@0X6F:8


    // set up the IRQ6 vector in RAM
    *(long *)0xEC2C = (long)ContextSwitch;


    // setup vector for switching outside of irq
    // Vector[xxx]in ram = ScheduleSlice;


/////////////////////////////// be sure and enable interrupts someplace!!


    // construct a new task - idle and fill it in
    idle = TaskCreate(IdleTask, IDLE_TASK_STACK_SIZE, 0, TASK_PRIORITY_IDLE);

    if (idle)
    {
        // Manually hook up the idle task

        // this is the task to run
        CurrentContext = idle;
    }

    return idle;
}





// To start the scheduler,
void SchedulerStart(TaskT *first)
{
    // point the stack to the first->stackpointer
    // manually call the scheduler: pull the irq0 chain
    // The first task should then set up the pre-emptive scheduling
    // and scheduling will then be running

    // We should never get here - tasks should be switching back and forth
    // and the idle should run if nothing else does
}





// Called from running task
// Only applies for running task
void
sleep(CybornTickT ticks)
{
    extern void TimerWait(CybornTickT timeout);

    // remove from the ready list only if the number of ticks is non-zero
    if (ticks)
    {
        // put the task on the timer waiting Q
        TimerWait(ticks);
    }
    else  // ticks == zero
    {
        // a zero arg means to just let another task run
        ScheduleYield();
    }
}








