#include "cyborn.h"

/*
Context Frame


When the interrupt occurs, the CCR and PC are pushed on the stack

/////////////////////////////////////////////////////////////////////
Example:
stackpointer before = xx1000
stackpointer after  = xx0FFC
 ======
|  CCR |  xx0FFC   <---- SP after interrupt
 ======
|  PC  |  xx0FFD
|      |  xx0FFE
|      |  xx0FFF
 ======
/////////////////////////////////////////////////////////////////////


\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
Hard coded vector to hard coded monitor location:
IRQ0 vector 16 - 00 00 11 2c
      40:   00 00           00 00
      42:   11 2c           11 2c


The cybiko has a hard interrupt service routine with a soft vector. The hard
service is in ROM and cannot be changed. It points to a vector in RAM that can
be changed

Here is where the vector points: inside the ROM
    112c:	01 30 6d f0 	01 30 6d f0       stm.l er0-er3,@-sp
    1130:	01 20 6d f4 	01 20 6d f4       stm.l er4-er6,@-sp
    1134:	01 00 6b 02 	01 00 6b 02 ec 14 mov.l @0xec14:16,er2
    1138:	ec 14
    113a:	5d 20       	5d 20             jsr   @er2
    113c:	01 20 6d 76 	01 20 6d 76       ldm.l @sp+,er4-er6
    1140:	01 30 6d 73 	01 30 6d 73       ldm.l @sp+,er0-er3
    1144:	56 70       	56 70             rte

Change the vector at 0xFFEC14 to point to a desired ISR


\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
IRQ0 and the timer int look similar
Before the return from the routine, the stack is like this:

EXTRA    <----- SP after the JSR
PC2     - address of next instruction (00113c in the case of IRQ0)
PC1
PC0
er6-3    <----- SP after interrupt
er6-2
er6-1
er6-0
er5-3
er5-2
er5-1
er5-0
er4-3
er4-2
er4-1
er4-0
er3-3
er3-2
er2-1
er2-0
er1-3
er1-2
er1-1
er1-0
er0-3
er0-2
er0-1
er0-0
CCR
PC2     - where to start/resume execution
PC1
PC0
         <----- SP before interrupt


When the scheduler is running - this is what the stack will need to look like
so that the task switch will point to the stack at the same place for each
task that needs to be continued. The Scheduler will load the stack pointer for
the selected task and then do an rts which will pop the PC off the stack,
return to the monitor rom section, load the registers off the stack, do an
rte which will complete the context switch by restoring the flags and
program counter. Here is the problem to solve: what to do about the FIRST time
that this task should run? Answer: since the instruction is an rts, the only
thing that needs to be on the stack is the vector for the entry point and move
the stackpointer to the entrypoint on the stack.


EXTRA
PC2   - entry point for task
PC1
PC0

This is done in an interrupt context by the scheduler. The scheduler doesn't
care about the registers, but the interrupt context needs to be adjusted
to be enabled at the beginning of each task. After the task has been properly
and interrupts enabled, the context switch where it does ALL of the registers
including CCR will be the normal activity.


*/





#define TIMING_SOURCE_WATCHDOG
#define CYBIKO_SYSTEM_CLOCK         11059200

// for changing the watchdog register
// The watchdog counts from 0 to 255 and then causes interrupt
// Interrupt freq = sysclock / DIV / 256
//                                          //    Hz       period
#define WATCHDOG_DIV_2              0x00    // 21600        46uS
#define WATCHDOG_DIV_64             0x01    //   675        1.5mS
#define WATCHDOG_DIV_128            0x02    //   337.5      3.0mS
#define WATCHDOG_DIV_512            0x03    //    84.375
#define WATCHDOG_DIV_2048           0x04    //    21.09375
#define WATCHDOG_DIV_8192           0x05    //     5.2734375
#define WATCHDOG_DIV_32768          0x06    //     1.318359375
#define WATCHDOG_DIV_131072         0x07    //     0.32958984375

#define WATCHDOG_DIV_MASK           0x07

#define WATCHDOG_DIV                WATCHDOG_DIV_64


//                                          clocks per timer tick
#define TIMER0_DIV_0              0x00    // stopped
#define TIMER0_DIV_8              0x01    // 1382400
#define TIMER0_DIV_64             0x02    //  172800
#define TIMER0_DIV_8192           0x03    //    1350

#define TIMER0_DIV_MASK           0x07

#define TIMER0_DIV                TIMER0_DIV_64



#define TICKS_PER_SECOND            (800)
#define SECONDS_TO_TICKS(s)         ((s) * TICKS_PER_SECOND)
#define MILLISECONDS_TO_TICKS(ms)   ((((ms) * TICKS_PER_SECOND) + 500) / 1000)





// storage for one pointer that keeps track of tasks blocked on a timer
//static TaskHeadT TaskSuspendList;


///////////////////////////////////////////////////////////// PRIORITIES
// These queues are for priorities and scheduling

typedef struct pq
{
    struct task *list;      // just point to the task on the list
    struct task *running;   // which task is current for this priority

    // if these point back to self, then it is the last one
    struct pq   *lower;     // next priority q that is a lower priority
    struct pq   *higher;    // next priority q that is a higher priority
} PriorityQueueT;




typedef void (*IsrEventT)(void *parameter);


static  PriorityQueueT pq[TASK_PRIORITIES];


// point this to the current queue of ready tasks - may have to lock this if
// getting it outside of the system tick - like during blocking
static PriorityQueueT *CurrentQueue;
static TaskT *CurrentTask;




/////////////////////////////////////////////////////////////   PERIODIC
/////////////////////////////////////////////////////////////   EVENT
#ifdef USE_PERIODICEVENT

// Periodic Event - checked at every event
typedef struct
{
    IsrEventT       event;
    void            *param;
    CybornTickT     period;
    CybornTickT     timer;
} PeriodicEventT;


// ONCHIP...
static PeriodicEventT *PeriodicEvent;


// Use like this:
// PeriodicEvent(dothis, MILLISECONDS_TO_TICKS(80));
// PeriodicEvent = event;
// -- to stop: PeriodicEvent = 0;
// API

// clear by setting event to null
// ie PeriodicEventSet(0, 0, 0);
void
PeriodicEventSet(IsrEventT event, void *parameter, CybornTickT period)
{
    static PeriodicEventT event;

    // doing it this way means we don't have to disable ints...
    PeriodicEvent = 0;

    // fill in the struct
    event.event     = event;
    event.param     = parameter;
    event.period    = period;
    event.timer     = 0;

    // point to ready-to-go struct
    PeriodicEvent = &event;
}

#endif
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////



// Bottom half processing can be done here in immediate tasks -
// The scheduling is turned off - the timer is stopped - to do these tasks

typedef struct bottomhalf
{
    struct bottomhalf *next;
    EntryT entry;
} BottomHalfT;



static volatile BottomHalfT *ImmediateExecution;


// IRQ Context
void inline
BottomHalfDo(BottomHalfT *bh)
{
    // add to the head of the list
    bh->next = ImmediateExecution;

    // add to the empty list
    ImmediateExecution = bh;

    SignalIsrSend(&BottomHalfWaiting);
}



// Top priority
void
BottomHalfRun(void)
{
    BottomHalfT *bh;


    for (;;)
    {
        // COUNTING SEMAPHORE
        // stay on the waiting list until signalled
        SignalWait(&BottomHalfWaiting);

        /////// disable interrupts

        // take an entry
        bh = ImmediateExecution;

        // take it off the list
        ImmediateExecution = ImmediateExecution->next;

        /////// enable interrupts

        // run the bh
        bh->entry();
    }
}







//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////


void inline
ScheduleDisable(void)
{
    // disable timer interrupt
}



void inline
ScheduleEnable(void)
{
    // enable timer interrupt -
}





CybornTickT CybornTicksPerSecond(void)
{
    return TICKS_PER_SECOND;
}




//////////////////////////////////////////////////////////////////////////////
// Call this with the task that needs to be added to the prioritized ready list
// DO NOT CALL from timer slice! Rewrite it there for speed, etc
// DO NOT ADD A TASK TWICE!
// Returns true if the task added had a higher priority than current
// The CurrentQueue is changed
void
ScheduleAdd(TaskT *task)
{
    PriorityQueueT *queue;
    TaskT *index;


    // REMOVE task from other blocking list
    task->prev->next = task->next;
    task->next->prev = task->prev;

    // ADD task to ready to run list

    // get the pointer to the list of which priority the task is on
    queue = (PriorityQueueT *)task->schedule;

    // if empty list, need to add task and reinsert queue
    if (queue->list == 0)
    {
        PriorityQueueT found;

        // link/point the task to itself
        task->next = task;
        task->prev = task;

        // The queue is not being used - set it all up
        // point the task list to this task
        queue->list = task;
        queue->running = task;

        // insert the queue into the other priority queues
        // there may only be one - note that this has to search and can take
        // time. But, it only has to do so when a queue was empty and now
        // has an entry. Further additions are not expensive. The bad case is
        // when a single task for this priority runs - waits - runs - waits.
        // The expense is offset, because it runs without having to let any
        // other tasks have a slice

        // FIND priority that is lower and insert new task before it
        // use the lowest and then use the lower - it will wrap to highest
        found = pq[TASK_PRIORITES-1].lower;

        for (;;)
        {
            if (queue->priority < found->priority)
            {
                // INSERT into double linked list
                queue->lower = found;
                queue->higher = found->higher;
                found->higher->lower = queue;
                found->higher = queue;

                break;
            }

            // look at the next lower queue
            found = found->lower;
        }
    }
    else  // list was not empty
    {
        // insert the new task before the one that is running
        // add to linked list of ready tasks for this priority
        // insert the new task after the index
        index = queue->running->prev;
        task->next          = index->next;
        index->next         = task;
        task->prev          = index;
        task->next->prev    = task;
    }

}






//////////////////////////////////////////////////////////////////////////////
// Call this with the task that needs to be added to the prioritized ready list
void
ScheduleIsrWake(TaskT *task)
{
    TaskT *index;

    if (task->priority < CurrentQueue->running->priority)
    {
        ScheduleAdd(task);
        _ContextSwitch();
    }
    else    // no need for a context switch - no priority change
    {
        ScheduleAdd(task);
    }

}




//////////////////////////////////////////////////////////////////////////////
// Call this with the task that needs to be added to the prioritized ready list
// DO NOT CALL from timer slice! Rewrite it there for speed, etc
// DO NOT ADD A TASK TWICE!
inline void
ScheduleWake(TaskT *task)
{
    if (task->priority < CurrentQueue->running->priority)
    {
        ScheduleDisable();
        ScheduleAdd(task);
        ScheduleEnable();

        // there is a difference in the priorities
        ScheduleSwitch();
    }
    else
    {
        ScheduleDisable();
        ScheduleAdd(task);
        ScheduleEnable();
    }

}




// Call before a task can run to register and set some internal data
// fill in the schedule infomation in the task header for the task
void
ScheduleRegister(TaskT *task)
{
    switch (task->priority)
    {
        // unknown priority gets LOW priority
        default:
        case TASK_PRIORITY_LOW:
            task->schedule = &PriorityLow;
            break;

        case TASK_PRIORITY_MED:
            task->schedule = &PriorityMed;
            break;

        case TASK_PRIORITY_HIGH:
            task->schedule = &PriorityHigh;
            break;
    }
}




// Moves a task from the running list it is on, to another list
// A task must be inserted in a list someplace!
inline void
ScheduleRemove(void)
{
    TaskT *task = CurrentQueue->running;

    // REMOVE task from running list and move it to another
    // This task will not be run again

    // since the task that was removed is the highest, another needs to be
    // chosen to run


    // list have only one entry? - then unlink this queue from the others
    // and switch to another queue - the queue will be lower since the
    // task that is being removed now should be the highest
    if (task->next == task->prev)
    {
        // remove this list from the higher and lower links
        CurrentQueue->higher->lower = CurrentQueue->lower;
        CurrentQueue->lower->higher = CurrentQueue->higher;

        // setup the new task to be run - can ALWAYS go lower - idle task runs
        CurrentQueue = CurrentQueue->lower;

        // register/indicate that queue has been removed and is empty
        CurrentQueue->list = 0;
    }
    else  // list has more than one entry
    {
        // check for the head of the list getting removed
        if (CurrentQueue->list == task)
        {
            // point to the next entry
            CurrentQueue->list = task->next;
        }

        // next to run is the next in the list
        CurrentQueue->running = task->next

        // Unlink the task by linking the entries around it to each other
        task->prev->next = task->next;
        task->next->prev = task->prev;
    }

}



// Moves the current task from the running list to another list
// A task must be inserted in a list someplace!
void
ScheduleWait(TaskT *index)
{
    TaskT   *task;

    // Don't let other scheduling go on
    ScheduleDisable();

    // save task and its pointers for relinking
    task = CurrentQueue->running;

    // REMOVE this task from running list - a new one will be scheduled
    // This task will not be run again
    ScheduleRemove();

    // Now transfer the task to another list
    // Add the task AFTER the given index

    task->next = index->next;
    index->next = task;
    task->prev = index;
    task->next->prev = task;

    // allow scheduling again
    ScheduleEnable();

    // change the context - the task in CurrentQueue will run
    ScheduleSwitch();

}




///////////////////////////////////////////////////////////////////////////////
// Deletes a task from the running list
void
ScheduleDelete(TaskT *task)
{
    // Don't let other scheduling go on
    ScheduleDisable();

    // Unlink the task by linking the entries around it to each other
    ScheduleRemove(task);

    // allow time scheduling
    ScheduleEnable();
}



// Puts the task on the suspend list
// NOTE: Any task put here will not be run again until put on the ready Q
// A task could suspend itself... but cant restart itself
//   ScheduleSuspend(task);
//  later: by another task
//   ScheduleNow(task);

#if 0
void
ScheduleSuspend(TaskT *task)
{
    TaskT *index;

    index = (TaskT *)&TaskSuspendList);

    // remove task from the ready list and send to the suspend list
    ScheduleRemove(task, index);
}
#endif



// Always ISR context!!! Will not be interrupted here.
// No need to turn off the interrupts in here - they are already off.

// Figures out what has gone on and which task should run next
void ScheduleIsrYield(void)
{
    // next task
    ScheduleNext();

    // should be inline function
    _ContextSwitch();
}



// also called as ScheduleSlice - just yields to the next process to run
inline void
ScheduleYield(void)
{
    // called inline by code to yield to next same priority task to run
    ScheduleNext();

    // do the context switch to begin the next task
    ScheduleSwitch();
}




// ISR Context function for doing context switch
// INIT: load the address of this function into vector for IRQ
// USE:  Cause the interrupt to jump here and do the context switch
//       or jump here from IRQ
// NOTE: make sure that this is done inline without using stack space and time
inline void
_ContextSwitch(void)
{
    // Current task needs to have its context saved -registers are on the stack
    CurrentTask->stackpointer = ER7;

    // load the stackframe of the task to be run - the new context
    CurrentTask = CurrentQueue->running;
    ER7 = CurrentTask->stackpointer;

    // return from exception which causes the context switch
    rte;
}



inline void
ScheduleSwitch()
{
    // Cause IRQ that will push context onto stack (by nature of IRQ)
    // register something=0xXX - causes interrupt that will vector to
    // ContextSwitch function
}




// At one time I thought it might be faster to check if a switch needs to be
// done. This would make a single task running faster because a switch would
// not have to be made since the next task would be the same task. The overhead
// is not worth it for the single case.

inline void
ScheduleNext(void)
{
    // next to run the head? Cant run the head, use head->next
    if (CurrentQueue->running->next == CurrentQueue)
    {
        // then use the first entry in queue (could be the same one)
        CurrentQueue->running = CurrentQueue->next;
    }
    else  // next entry is not the head, assign it
    {
        // choose the next task to be run based on next task in this queue
        CurrentQueue->running = CurrentQueue->running->next;
    }
}




// At one time I thought it might be faster to check if a switch needs to be
// done. This would make a single task running faster because a switch would
// not have to be made since the next task would be the same task. The overhead
// is not worth it for the single case. Or NOT! heh

inline void
ScheduleNext2(void)
{
    // just make the next task in the queue the one that is running
    CurrentQueue->running = CurrentQueue->running->next;
}




// what should the idle task do?
// system stats?
// Its the first one to start so do initialization
static void IdleTask(void *nothing)
{
    // Do things here to initialize
    // Most important is that the stack for this task is filled with right vals
    // 1> Set Watchdog to clock divide and repeat output
    // 2> Set Watchdog vector in onchip RAM to point to TimeScheduler
    // 3> Enable the Watchdog timer interrupt - next interrupt will
    //    schedule the high priority task - if there is one

    for (;;)
    {
        // Do something idle here - never sleep or be taken off the ready Q
        // Check for screensaver?
        // Should we power down?
        // Preemption will break in here and choose a task with higher priority

    }
}


#if 0
// Example starting
TaskT *begin;

// set up the list for tasks
begin = SchedulerInit();
if (begin)
{
    // start all of the tasks that we want
    // NewTask.... ScheduleAdd.... repeat

    SchedulerStart(begin);

#endif



#define TASK_STACK_SIZE_MIN  32

#define IDLE_TASK_STACK_SIZE 68


// ISRs are off
// Initializes the scheduler with an idle task - must have at least one task
// Returns pointer to idle task - or 0 if error - the scheduler is ready to run
TaskT *SchedulerInit(void)
{
    static TaskT *idle;  // static cause it must persist
    int i;


    // setup vector for switching outside of irq
    // Vector[xxx]in ram = _ContextSwitch;


    // initialize the priority queues
    for (i = 0; i<TASK_PRIORITIES; i++)
    {
        pq[i].next = 0;
    }


    // construct a new task
    idle = TaskCreate(IdleTask, IDLE_TASK_STACK_SIZE, 0, TASK_PRIORITY_LOW);

    if (idle)
    {
        // manually add to the low priority queue
        pq[TASK_PRIORITY_LOW].priority = TASK_PRIORITY_LOW;
        pq[TASK_PRIORITY_LOW].lower = pq[TASK_PRIORITIES-1];
        pq[TASK_PRIORITY_LOW].higher = pq[TASK_PRIORITIES-1];

        // etc...
    }

    return idle;
}






// To start the scheduler,
void SchedulerStart(TaskT *first)
{
    // point the stack to the first->stackpointer
    // manually call the scheduler: pull the irq0 chain
    // The first task should then set up the pre-emptive scheduling
    // and scheduling will then be running

    // We should never get here - tasks should be switching back and forth
    // and the idle should run if nothing else does
}



// Called from running task
// Only applies for running task
void
sleep(CybornTickT ticks)
{
    // remove from the ready list only if the number of ticks is non-zero
    if (ticks)
    {
        // put the task on the timer waiting Q
        TimerBlock(ticks);
    }
    else  // ticks == zero - a zero arg means to just let another task run
    {
        // let another same-priority task run
        ScheduleYield();
    }
}


