#include "cyborn.h"

#include "stdlib.h"




// The number of combined semaphores, and mutexes
// Queues use a flag internally!
#ifndef FLAGS_MAX
#define FLAGS_MAX 40    // default if not overridden some place else
#endif



// All about TAS instruction
/*
The TAS instruction does 2 things:
1> Tests the byte pointed to with 0 and sets the CCR accordingly
2> Sets bit 7 of the byte to 1
The operation cannot be divided and will not be interrupted. This instruction
is used for syncronization.

before  after       ZFlag   NFlag
0x00    0x80        set     clr     ; flag was zero - now it is taken
0x80    0x80        clr     set     ; flag was taken

0x84    0x84        clr     set     ; flag was taken
0x04    0x84        clr     clr     ; flag was taken
*/

extern unsigned char _atomictake(unsigned char *flag);




//////////////////////////////////////////////////////////////////////////
/////////  P R I V A T E   D A T A
// onchip
static volatile unsigned char FlagPool[FLAGS_MAX];
static int NextFlag;




CybornStatusT
CyqInit(CybornQueueT *cyq, unsigned long size)
{
    if (size)
    {
        cyq->queue = (QueueT *)malloc(size);
        if (cyq->queue)
        {
            CybornStatusT status;

            // point to the front of the queue
            cyq->head = cyq->queue;
            cyq->next = cyq->queue;
            cyq->end  = cyq->queue + size;
            status = SignalInit(&cyq->sem, 0);
            if (status != CYBORN_STATUS_OK)
            {
                // free the memory and report a problem with resource
                free(cyq->queue);
                return CYBORN_STATUS_RESOURCE;
            }
        }
        else
        {
            // no memory
            return CYBORN_STATUS_MEMORY;
        }
    }
    else
    {
        // the size is zero - no good...
        return CYBORN_STATUS_EMPTY;
    }

    return CYBORN_STATUS_OK;
}




void
CyqDestroy(CybornQueueT *cyq)
{
    if (cyq)
    {
        SignalDestroy(&cyq->sem);
        free(cyq->queue);
    }
}


// Notes:
//  Many tasks can be waiting on characters - these are not full messages

CybornStatusT
CyqWait(    CybornQueueT    *cyq,       // which queue to wait on
            char const      *message,   // where to put message
            unsigned long   *count)     // size of message
{
    // wait for a semaphore that a message has been posted to the queue
    SignalWait(&cyq->sem);

    // TODO: LOCK the access to the queue pointers - turn off ISRs!

    // if there is a message - copy it in -
    *count = cyq->next - cyq->queue;
    memcpy(message, cyq->queue, *count);

    // reset the pointer to the front of the queue
    cyq->next = cyq->queue;

    // TODO: UNLOCK

    return ST_OK;
}




//
// Example: transmits characters from the queue to the RS232 port
// : wait for char(s), for each char: wait for TXready signal, send char

// System Consumer Task:
// char buffer[MESS_MAX];
// char *message;
//
// for (;;)
// {
//      // point the message to the buffer
//      message = buffer;
//      status = CyqWait(RS232Transmit, message, &size);
//      check status - ok? go on - bad? retry...
//      while (size--)
//      {
//          //Is the tx buffer ready?
//          SignalWait(TxBufferEmpty);
//
//          //put the message in the byte to send
//          TXOUTREGISTER(*message++);
//      }
// }


// SEND CHARACTERS out the RS232
// ISR Context: when Tx is done - this is called to flag as ready
// TxDoneAndReady
//      SignalGive(TxBufferEmpty);   // ok in ISR
//      rte


// From program producer task:
// NOTE: this could be a higher level function: TxRS232(char);
//   The above queue could be setup in a task that waits
// RS232Transmit = CyqInit(&Q, size);
// This will send the char to a message queue that will transmit over RS-232
// spin on sending - use sleep if the baud rate is slow (<9600)
// while (CYBORN_STATUS_FULL == CyqAdd(RS232Transmit, char2send)) ;


// Character Queue
// CyqAdd works like this:
// Add another entry on the queue - signal that queue has entries
// If all entries are filled, CyqAdd exits with an error code
// ISR safe
CybornStatusT
CyqAdd(CybornQueueT *queue, char character)
{
    if (queue->next != queue->end)
    {
        // dont use a mutex here because a context change is too much
        // overhead for just a small bit of code, but dont turn off interrupts

        *queue->next++ = character;

        // indicate that this queue has data - use a binary sem
        SignalGive(&queue->sem);
    }
    else
    {
        return ST_FULL;
    }

    return ST_OK;
}







//////////////////////////////////////////////////////////////////////////
//////////////  M U T E X E S
//////////////////////////////////////////////////////////////////////////
// Don't use mutexes in ISRs.
// Use mutexes to guard against two (or more) tasks accessing code path at
// the same time.

#if 0
    .global __semgrab

    ;er0 should point to a nice place in onchip RAM so it is fast

    ;er0 will have the pointer to the semaphore
    ;if the semaphore WAS empty then the carry flag will be clear
    ;the semaphore is ALWAYS taken (non-zero) after 'tas'
    ;Zero Flag if Sem was:- Empty: 1, Taken: 0
    ;Neg Flag if Sem was:-  Empty: 0, Taken: 1
    tas @er0        ;test and set
    bmi taken       ;if taken - zero flag would be 0 - neg flag would be 1
    sub er0, er0    ;else it was empty - clear er0 (return value)
taken:
    rts


#endif




// Try to see if the mutex has been taken - take it if we can
CybornStatusT
CriticalTry(CybornMutexT *mutex)
{
    // do an atomic grab of a semaphore - in asm? - inline - macro
    // returns a zero if it was empty
    if (_atomictake(mutex->flag) == 0)
    {
        // got the semaphore - we are now the owner
        mutex->owner = CurrentTask;

        return ST_OK;
    }

    // couldn't get the mutex - still blocked
    return ST_BLOCKED;
}



// DO NOT USE THIS IN ANY ISR!!
void inline
CriticalEnter(CybornMutexT *mutex)
{
    // do an atomic grab of a semaphore - in asm? - inline - macro
    // returns zero if the flag was already taken
    // Uses the TAS instruction. The zero code is set if the byte was zero
    //   - zero means that the byte was not taken. After the TAS, the byte
    //     bit 7 is always set.
    // The take must look like this -
    //      TAS flag
    //      jump on zero (zero code is set) to 'got the flag'
    // atomictake:
    //      tas @er0;
    //      mov ccr, r0;
    if (_atomictake(mutex->flag))
    {
        ////  W A I T  ////
        ScheduleBlock(mutex->next);

        // We will return when we get the mutex and get time to run
        // ...
        //    ...
        //        Now running - only can happen if we are the owner
    }
}



// Notes:
//  One task can own/take many mutexes  - mutex has to track owner
//  Task can only block on only one mutex - task can track blocked list
//  Mutex can be owned and entered by only one task at a time
//  NEVER used in ISRs
void inline
CriticalExit(CybornMutexT *mutex)
{
    // If there is another task waiting to get in - wake it
    // If the next in list is self, then there is no NEW owner
    if (mutex->next != (TaskT *)&mutex->next)
    {
        // Make the new ready to run
        // The owner will be moved from the waiting list to the ready list
        // If the task has a higher priority, it will switch now
        ScheduleWake(mutex->next);
    }
}



// This will reset and use a mutex
CybornStatusT
CriticalInit(CybornMutexT *mutex)
{
    int endsem;

    // sanity check
    if (mutex)
    {
        // clear the owner
        mutex->owner = TASK_NONE;

        // set the waiting list to point to self and wrap the next and prev
        mutex->next = (TaskT *)&mutex->next;
        mutex->prev = (TaskT *)&mutex->next;

        // can we recycle the flag coming in?
        if (*mutex->flag & 0x7F)
        {
            // reuse this flag - as not taken
            *mutex->flag = 0;
        }
        else
        {
            // set the endpoint for wrap around
            endsem = NextFlag;

            for (;;)
            {
                if (FlagPool[NextFlag] == 0)
                {
                    FlagPool[NextFlag] = NextFlag;
                    mutex->flag = &FlagPool[NextFlag];
                    return CYBORN_STATUS_OK;
                }

                // check for wrap around
                if (++NextFlag >= FLAGS_MAX)
                {
                    NextFlag = 0;
                }

                // have we checked all of the sems?
                if (NextFlag == endsem)
                {
                    break;
                }

            }

            return CYBORN_STATUS_EMPTY;
        }

        return CYBORN_STATUS_OK;
    }

    return CYBORN_STATUS_NULL;
}



//////////////////////////////////////////////////////////////////////////
//////////////  S E M A P H O R E S
//////////////////////////////////////////////////////////////////////////
// TODO: Add other types of semaphores - allocate semaphores completely in
//       onchip RAM
// Notes:
//  Semaphore is used to signal event has happened - this is a Binary sem
//  Example: TaskA waits for a semaphore so it can process something
//    TaskB raises the semaphore when it has made data ready
//
//  This is a binary semaphore - taking once clears it even if it has been
//    given many times.
//
//  There are no owners of semaphores just givers and takers - if it is taken
//  it is not owned. It is just not 'there' anymore (it was taken!)
//
// ISR cannot give this
void
Signal(CybornSemT *semaphore)
{
    // turn off all interrupts
    INTERRUPTS_DISABLE;

    if (semaphore->flag < semaphore->max)
    {
        // Any task waiting for this signal?
        if (semaphore->next != (TaskT *)&semaphore->next)
        {
            ScheduleWake(semaphore->next);
        }
        else  // not another task waiting so save the flag
        {
            semaphore->flag++;
        }
    }
    
    INTERRUPTS_ENABLE;

}





// Interrupts are off - only use in interrupt context
void
SignalIsr(CybornSemT *semaphore)
{
    // Any task waiting for this signal?
    if (semaphore->next != (TaskT *)&semaphore->next)
    {
        TaskT *task = (TaskT*)semaphore->next;

        // unlink
        LIST_UNLINK(task);

        // later - if there is a timer event associated, cancel it
        // timerRemove(task);

        // ADD task to ready to run list
        if (task->priority == TASK_PRIORITY_HIGH)
        {
            // Add to the FRONT of the ready list
            LIST_LINK_FRONT(task, pTaskReadyList)

            // need to switch to new task if priority is higher
            if (CurrentContext->priority > TASK_PRIORITY_HIGH)
            {
                // assign the task
                if (SwitchTask && (SwitchTask->priority > task->priority()
                {
                    SwitchTask = task;
                }
            }
        }
        else
        {
            // Add to the BACK of the ready list
            LIST_LINK_BACK(task, pTaskReadyList)
        }

    }
    else  // not another task waiting so save the flag
    {
        // cannot give the flag more than once just set it
        semaphore->flag = 1;
    }

}



// DO NOT USE THIS IN ISR!!
// May not wait at all if the semaphore is available
void
SignalWait(CybornSemT *semaphore)
{
    INTERRUPTS_DISABLE;

    // available?
    if (semaphore->flag)
    {
        // take it
        semaphore->flag--;
        INTERRUPTS_ENABLE;

    }
    else  // no semaphore - task must block
    {
        INTERRUPTS_ENABLE;

        ScheduleBlock(semaphore->next);
    }

    // RETURN here when we get the signal given
    return;
}




// Will not wait if the semaphore is not available - takes it if it can
CybornStatusT
SignalTry(CybornSemT *semaphore)
{
    INTERRUPTS_DISABLE;
    if (semaphore->flag)
    {
        // we'll take it and own it now
        semaphore->flag--;

        INTERRUPTS_ENABLE;
        return CYBORN_STATUS_OK;
    }

    INTERRUPTS_ENABLE();

    // not given yet
    return CYBORN_STATUS_BLOCKED;
}




// Can create the binary semaphore as taken or not

CybornStatusT
SignalInit(CybornSemT *semaphore, int initial, SemTypeT type)
{
    // get a flag
    if (NextFlag < FLAGS_MAX)
    {
        // get the pointer to
        semaphore->flag = &FlagPool[NextFlag++];
        *semaphore->flag = initial;

        // no waiters, but use a fake spot in the semaphore to look like a list
        semaphore->next = (TaskT *)&semaphore->next;
        semaphore->prev = (TaskT *)&semaphore->next;
        return CYBORN_STATUS_OK;
    }
    else
    {
        return CYBORN_STATUS_RESOURCE;
    }
}




// timed signals.... ?
#if 0
// DO NOT USE THIS IN ISR!!
// May not wait at all if the semaphore is available
CybornStatusT
SignalWait(CybornSignalT *signal, CybornTicksT timeout)
{
    // available?
    if (signal->flag)
    {
        // take it
        signal->flag = 0;
    }
    else  // not available - must wait - no longer ready
    {
        if (ScheduleWait(signal->next, timeout) == CYBORN_STATUS_TIMEOUT)
        {
            // didn't get the flag; the timer kicked us out
            return CYBORN_STATUS_TIMEOUT;
        }
    }

    return CYBORN_STATUS_OK;
}



void
SignalGive(CybornSignalT *signal)
{
    // only give the flag if it has not been given before!
    if (signal->count < 0)
    {
        // inc - now is it a zero?
        if (++signal->count == 0)
        {
            // if a task is waiting for this signal add it to the ready list
            if (signal->waiting)
            {
                // Put the waiting task on the ready Q, if higher priority
                // it will run.
                ScheduleWake(signal->waiting);
            }
        }
    }
}

#endif


