

#include "cyborn.h"



// Timing
// Two kinds of timing here:
//  1) Time accurate - for sleep, process slicing
//  2) Time periodic - doesn't matter the exact time, just that it happens

// The time accurate can be done at 1ms increments to be more precise
// The time slicer can be the watchdog that runs all the time to switch
//  tasks whenever the watchdog times out. Watchdog timer around 675 Hz.

// But the watchdog is harder to turn off and on.


extern TaskListT TaskReadyList;
extern TaskListT *pTaskReadyList;

extern TaskT *CurrentContext;
extern TaskT *NextContext;


static TaskListT BlockedOnTicks;

// CYBIKO_CLOCK_HZ


typedef enum
{
    WATCHDOG_FREQ_0x00 = 0x00,  //  clk/2       11592000 / 256 / 2
    WATCHDOG_FREQ_0x01,         //  clk/64      11592000 / 256 / 64
    WATCHDOG_FREQ_0x02,         //  clk/128         etc
    WATCHDOG_FREQ_0x03,         //  clk/512
    WATCHDOG_FREQ_0x04,         //  clk/2048
    WATCHDOG_FREQ_0x05,         //  clk/8192
    WATCHDOG_FREQ_0x06,         //  clk/32768
    WATCHDOG_FREQ_0x07,         //  clk/131072
} WatchdogFreqT;


struct
{
    unsigned short  period;
    void            *parameter;
    EntryT          entry;
    CybornTickT     count;
} PeriodicEntry;


struct
{
    void            *parameter;
    EntryT          entry;
} WatchdogEntry;




void
PeriodicAdd(EntryT      entry,          // called upon timeout
            void        *parameter,     // parameter to send to callback
            CybornTickT count,          // how many times? 0: always
            CybornTickT delay,          // ticks to wait before starting
            CybornTickT period)         // ticks between callbacks
{
    PeriodicEntry.entry     = entry;
    PeriodicEntry.period    = period;
    PeriodicEntry.count     = count;
    PeriodicEntry.parameter = parameter;

#if 0
    // setup and enable the interrupt
    timer0->counter = delay;

    // enable
    timer0->tsts.enable = 1;
#endif
}





// assign a task to be executed periodically at a high priority
void
SystemPeriodicEvent(   EntryT      entry,
               void        *parameter)
{
    WatchdogEntry.entry = entry;
    WatchdogEntry.parameter = parameter;
}




#if 0

// Remove the timer from the list
// Does not put it on another list
CybornStatusT
TimerRemove(TaskT *task)
{
    // save/add ticks to next timer entry as adjustment
    if (task->tnext)
    {
        task->tnext->ticks += task->ticks;
        task->tnext->tprev = task->tprev;
    }

    // unhook 'next' from timer list
    task->tprev->next = task->tnext;

    // not on the list
    return CYBORN_STATUS_OK;
}

#endif



// ISR context - called from an ISR
// make all tasks ready that have spent their time sleeping
static void
TimerCheck(void)
{
    register TaskT  *index;

    // search for any tasks that should be scheduled as ready
    // start at the beginning
    index = BlockedOnTicks.next;

    // look for timers that have expired
    while (index != BlockedOnTicks.prev)
    {
        // remove all of the zero time tasks from the waiting list
        // and make them ready by moving to the ready list
        if (index->ticks == 0)
        {
            // Ready! Put on the ready list
            // REMOVE task from the timer blocking list
            LIST_UNLINK(index);

            // ADD task to ready to run list
            if (index->priority == TASK_PRIORITY_HIGH)
            {
                // Add to the FRONT of the ready list
                LIST_LINK_FRONT(index, pTaskReadyList->next);
            }
            else  // low - lower priority
            {
                // Add to the BACK of the list
                LIST_LINK_BACK(index, pTaskReadyList->next);
            }

            // go to the next entry - remove next zero or dec first non 0
            index = index->next;
        }
        else   // no ready event
        {
            // decrement the first non-zero counter and leave
            index->ticks--;
            break;
        }
    }

    return;
}




// could be called from 'sleep'
// BlockedOnTicks is list of tasks blocked on ticks

//////////////////////////////////////////////////////////////////////////
//  Add Time Blocked task
//////////////////////////////////////////////////////////////////////////
// Uses lots of time here to put the task in the right spot, but the
// payoff is that checking timeout is very simple. The first one on the
// list is quick. The tasks that have a longer timeout use more time to
// get on the list. Seems good.
void
TimerWait(CybornTickT timeout)
{
    TaskT *index;

    INTERRUPTS_DISABLE;

    index = BlockedOnTicks.next;

    // only insert an entry here if the list is not empty
    if (index != (TaskT *)&BlockedOnTicks.next)
    {
        // Check if new entry time is later by comparing to ticks in entry
        // If the new time is longer,it must be put behind others
        if (timeout > index->ticks)
        {
            // everything is stored as a delta time
            // start with absolute and go down from there

            // MUST insert this event after the head - search for it
            for (;;)
            {
                // remove the ticks that will pass before timeout
                timeout -= index->ticks;

                // end of the list? Does the end point here?
                if (index->next == (TaskT *)&BlockedOnTicks.next)
                {
                    // insert after index at the end of the list
                    break;
                }

                // was not the end - does the timeout fit in here?
                if (timeout < index->next->ticks)
                {
                    // remove the time from the next because its relative
                    index->next->ticks -= timeout;

                    // done looking - insert after index
                    break;
                }

                // go to next timer entry
                index = index->next;
            }
        }
        else  // same or less - put the current context at the front
        {
            // readjust the delta  NOTE: this could result in a zero
            index->ticks -= timeout;
        }
    }

    // assign the timeout
    CurrentContext->ticks = timeout;

    /////////////////////////////////////////////////////////////////////////
    //////////////////////////////////// Change the context /////////////////
    // Change the context
    NextContext = CurrentContext->next;

    // current task is removed from the ready list and inserted at the index
    // REMOVE CurrentContext from the ready list
    LIST_UNLINK(CurrentContext);

    // insert the task into the list to wait for its time to come up
    LIST_LINK_FRONT(CurrentContext, index);

    // if the next task priority is lower, go to the first (hi pri) task
    if (IS_LOWER_THAN_CURRENT(NextContext))
    {
        // point to the head of the ready task list
        NextContext = pTaskReadyList->next;
    }

    // switch context
    CONTEXT_SWITCH;

    // enable interrupts - as soon as enabled - the IRQ6 should fire
    INTERRUPTS_ENABLE;

    /////////////////////////////////////////////////////////////////////////
}



// ISR for the timer interrupt
// Called every 50ms
// Do not use any local variables here because of task switching
// This was called like this:
// . Interrupt
// . Interrupt vector
// . push all
// . jsr PeriodicISR
void
PeriodicISR(void)
{
    extern void ContextSwitch(void);

    // keypress/release?
    //KeyboardCheck();  // move this to the periodic timer section

    // Schedule any timer blocked tasks
    TimerCheck();

    // Round robin scheduling
    // is the next priority too low? Is there something higher on the queue?
    if (CurrentContext->next->priority < pTaskReadyList->next->priority)
    {
        // switch to the head of the list - where the high priorities are
        NextContext = TaskReadyList.next;
    }
    else
    {
        // point to the next task - round robin style
        NextContext = CurrentContext->next;
    }

    //////////////////////////////////////////////////////////////////////
    // change the context

    // since the stack is already setup - just go to the switch
    ContextSwitch();
}

