


#include "cyborn.h"


#include "event.h"


#define MAKE_EVENT(x) ((x) << 24)


#define CONSOLE_MAX 7


// events
typedef enum
{
    EVENT_KEY_DOWN  = MAKE_EVENT(0x01),
    EVENT_RTC       = MAKE_EVENT(0x02),
    EVENT_REDRAW    = MAKE_EVENT(0x03),
    EVENT_TIMER     = MAKE_EVENT(0x04),
    EVENT_STORAGE   = MAKE_EVENT(0x05),
    EVENT_RF_IN     = MAKE_EVENT(0x06),
    EVENT_SHOW      = MAKE_EVENT(0x07),
    EVENT_HIDE      = MAKE_EVENT(0x08),
    EVENT_KEY_UP    = MAKE_EVENT(0x09),
    EVENT_EXIT      = MAKE_EVENT(0x0A),
} EventT;

typedef char ConsoleT;
#define APP_NAME_SIZE 16


// structure for an application
typedef struct app
{
    DisplayT    *display;                   // pointer to display memory
    char        name[APP_NAME_SIZE + 1];    // name of the application
    NoteT       *queue;                     // message queue
    ConsoleT    console;                    // which console associated with
    TaskT       *tasklist;                  // what tasks are associated
    struct app  *next;
} ApplicationT;



ApplicationT    *CurrentApplication;
ApplicationT    *ApplicationPrior;



// Send an event to an application
// The events can be aggregated or manipulated for efficency
// NOTE: is this level really needed?
void
EventSend(ApplicationT *app, EventT event)
{
    *app->queue.head++ = event;
    Signal(app->queue.sem);
}



// this is where the OS starts its life
// This is the only one in the system
void
CybornMain(void)
{
    // initialize the kernel functions:
    // this should take care of tasks, ipc and sync
    KernelInitialization();

    // file system setup - structures, etc
    FileSystemInitialization();


    // initialize the screen

    // setup the consoles

    // start task for watching the keyboard - high priority

    // setup the GUI

    // setup the RS-232 port

    // start the info task

    // Create the first application -
    SystemMonitor = (ApplicationT *)malloc(sizeof(ApplicationT));
    CurrentApplication = SystemMonitor;

    // Create the keyboard polling ISR
    // Setup the interval to around 100ms to poll keyboard
    TimerIsr(Keyboard);


}




#define COMMAND(x) ((x) & 0xFF000000)



// system monitor task
// Handles switching between consoles/apps and showing info
//
void
SystemMonitor(void)
{
    UINT16  key;
    NoteT   note;
    StatusT status;

    // The system loop
    for (;;)
    {
        // all keyboard input comes here - what else?
        status = NoteRead(OSQueue, &note);

        if (status != STATUS_OK) continue;

        // process the incoming note
        switch (COMMAND(note))
        {
            case EVENT_KEY_DOWN:
                // translate the pattern to ascii - only if asked for
                // key = ASCII(note & 0xFFFF);
                if (key == RAW_KEY_HELP)
                {
                    // show the system info application
                    //  it always exists - there is no interaction it is just info
                    //
                    if (CurrentApplication != SystemMonitor)
                    {
                        ApplicationPrior = CurrentApplication;

                        // changing the application now
                        // tell the current application it will not be shown
                        EventSend(CurrentApplication, EVENT_HIDE);

                        // how to know when the current app is hidden
                        CurrentApplication = SystemMonitor;
                        EventSend(CurrentApplication, EVENT_SHOW);
                    }
                }
                else if (key >= KEY_RAW_F1 && key <= KEY_RAW_F7)
                {
                    // events are sent async - the application may pull off the
                    // message at any time. The new one could get the message first
                    if (CurrentApplication != ApplicationList[key - KEY_F1])
                    {
                        EventSend(CurrentApplication, EVENT_HIDE);

                        CurrentApplication = ApplicationList[key - KEY_F1];
                        EventSend(CurrentApplication, EVENT_SHOW);
                    }
                }
                else
                {
                    // send the key to the app - send the raw
                    // let the application translate it in user space
                    EventSend(EventSend(CurrentApplication, note);
                }
                break;

            // the key was released - filter the special keys here
            case EVENT_KEY_UP:
                if (key == RAW_KEY_HELP)
                {
                    NoteSend(CurrentApplication->queue, EVENT_HIDE);

                    // restore the application that was there before
                    CurrentApplication = ApplicationPrior;
                    NoteSend(CurrentApplication->queue, EVENT_SHOW);
                }
                else if (key >= KEY_RAW_F1 && key <= KEY_RAW_F7)
                {
                    // do nothing here
                }
                else
                {
                    // send the key to the app - send the raw
                    // let the application translate it in user space
                    EventSend(EventSend(CurrentApplication, note);
                }

                break;

        }
    }
}



void
AppStart(void *arg)
{
    for (index = 1; index < CONSOLE_MAX; index++)
    {
        TaskT *task;


        id = malloc(sizeof(ApplicationT));
        ApplicationList[index] = id;

        // need a display to write
        // some apps dont need a display
        id->display = malloc(sizeof(DisplayT));

        // create task
        task = TaskCreate(AppDefault, TASK_PRIORITY_LOW);

        // register a task with the application
        id->tasklist = task;

        // queue for events
        id->queue = NoteCreate(20);

        // add application to application list
        id->next = ApplicationList;
        ApplicationList = id;

        // start things running
        TaskRun(task);
    }
}




// If no other app has been started, this is the one that is running
//  for each console
void
AppDefault(ApplicationT *id)
{
    ApplicationT *id;
    CybornStatusT status;


    // get registered with the OS
    // we get a filled in id
    status = CybornAttach(&id, APP_DISPLAY_ON, EVENT_MASK_KEYS | EVENT_MASK_TIMER);


    for (;;)
    {
        // wait here until a note comes to us
        status = NoteRead(queue, &note);

        switch (note & 0xFF000000)
        {
            case EVENT_SHOW: // do something - we are being displayed
                break;

            case EVENT_HIDE: // do nothing - we are no longer on display
                break;

            case EVENT_EXIT: // leave this app
                // release from OS
                CybornDetach(id);
                break;

        }

    }

}



// This runs in a periodic interrupt context about 300 times a second
// interrupts are off for some of the time
void
SystemPoll(void)
{
    // First check for any rs232 received data
    if (rs232_rx_head != rs232_rx_tail)
    {
        if (Rs232Handler)
        {
            // tell someone there are bytes ready to read
            EventSend(Rs232Handler, EVENT_RS232_RX);
        }
    }

    // turn interrupts on
    ENABLE_INTERRUPTS();



    // Timer Events
    // Go through timers
    if (timers->next)
    {
        timer = timers->next;

        while (timer)
        {
            if (timer->time)
            {
                if (--timer->time == 0)
                {
                    EventSend(timer->owner, EVENT_TIMER | timer->ID);

                    // if timer is resetable
                    if (timer->mode & TIMER_MODE_RESET)
                    {
                        timer->time = timer->resettime;
                    }
                }
            }
            timer = timer->next;
        }

    }
    
    // dont check the keyboard so often
    if (interruptcounter++ & 0x0f) return;

    key = keyboard();
    if (key != keyold)
    {
        if (key)
        {
            EventSend(CurrentTask->events, EVENT_KEY_DOWN | key);
        }
        else
        {
            EventSend(CurrentTask->events, EVENT_KEY_UP | keyold);
        }
    }


}





